
#ifndef TTKARRAYEDITOR_EXPORT_H
#define TTKARRAYEDITOR_EXPORT_H

#ifdef TTKARRAYEDITOR_STATIC_DEFINE
#  define TTKARRAYEDITOR_EXPORT
#  define TTKARRAYEDITOR_NO_EXPORT
#else
#  ifndef TTKARRAYEDITOR_EXPORT
#    ifdef ttkArrayEditor_EXPORTS
        /* We are building this library */
#      define TTKARRAYEDITOR_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKARRAYEDITOR_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKARRAYEDITOR_NO_EXPORT
#    define TTKARRAYEDITOR_NO_EXPORT 
#  endif
#endif

#ifndef TTKARRAYEDITOR_DEPRECATED
#  define TTKARRAYEDITOR_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKARRAYEDITOR_DEPRECATED_EXPORT
#  define TTKARRAYEDITOR_DEPRECATED_EXPORT TTKARRAYEDITOR_EXPORT TTKARRAYEDITOR_DEPRECATED
#endif

#ifndef TTKARRAYEDITOR_DEPRECATED_NO_EXPORT
#  define TTKARRAYEDITOR_DEPRECATED_NO_EXPORT TTKARRAYEDITOR_NO_EXPORT TTKARRAYEDITOR_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKARRAYEDITOR_NO_DEPRECATED
#    define TTKARRAYEDITOR_NO_DEPRECATED
#  endif
#endif

#endif /* TTKARRAYEDITOR_EXPORT_H */
