
#ifndef TTKFIBERSURFACE_EXPORT_H
#define TTKFIBERSURFACE_EXPORT_H

#ifdef TTKFIBERSURFACE_STATIC_DEFINE
#  define TTKFIBERSURFACE_EXPORT
#  define TTKFIBERSURFACE_NO_EXPORT
#else
#  ifndef TTKFIBERSURFACE_EXPORT
#    ifdef ttkFiberSurface_EXPORTS
        /* We are building this library */
#      define TTKFIBERSURFACE_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKFIBERSURFACE_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKFIBERSURFACE_NO_EXPORT
#    define TTKFIBERSURFACE_NO_EXPORT 
#  endif
#endif

#ifndef TTKFIBERSURFACE_DEPRECATED
#  define TTKFIBERSURFACE_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKFIBERSURFACE_DEPRECATED_EXPORT
#  define TTKFIBERSURFACE_DEPRECATED_EXPORT TTKFIBERSURFACE_EXPORT TTKFIBERSURFACE_DEPRECATED
#endif

#ifndef TTKFIBERSURFACE_DEPRECATED_NO_EXPORT
#  define TTKFIBERSURFACE_DEPRECATED_NO_EXPORT TTKFIBERSURFACE_NO_EXPORT TTKFIBERSURFACE_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKFIBERSURFACE_NO_DEPRECATED
#    define TTKFIBERSURFACE_NO_DEPRECATED
#  endif
#endif

#endif /* TTKFIBERSURFACE_EXPORT_H */
