
#ifndef TTKBLANK_EXPORT_H
#define TTKBLANK_EXPORT_H

#ifdef TTKBLANK_STATIC_DEFINE
#  define TTKBLANK_EXPORT
#  define TTKBLANK_NO_EXPORT
#else
#  ifndef TTKBLANK_EXPORT
#    ifdef ttkBlank_EXPORTS
        /* We are building this library */
#      define TTKBLANK_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKBLANK_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKBLANK_NO_EXPORT
#    define TTKBLANK_NO_EXPORT 
#  endif
#endif

#ifndef TTKBLANK_DEPRECATED
#  define TTKBLANK_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKBLANK_DEPRECATED_EXPORT
#  define TTKBLANK_DEPRECATED_EXPORT TTKBLANK_EXPORT TTKBLANK_DEPRECATED
#endif

#ifndef TTKBLANK_DEPRECATED_NO_EXPORT
#  define TTKBLANK_DEPRECATED_NO_EXPORT TTKBLANK_NO_EXPORT TTKBLANK_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKBLANK_NO_DEPRECATED
#    define TTKBLANK_NO_DEPRECATED
#  endif
#endif

#endif /* TTKBLANK_EXPORT_H */
