
#ifndef TTKUSERINTERFACEBASE_EXPORT_H
#define TTKUSERINTERFACEBASE_EXPORT_H

#ifdef TTKUSERINTERFACEBASE_STATIC_DEFINE
#  define TTKUSERINTERFACEBASE_EXPORT
#  define TTKUSERINTERFACEBASE_NO_EXPORT
#else
#  ifndef TTKUSERINTERFACEBASE_EXPORT
#    ifdef ttkUserInterfaceBase_EXPORTS
        /* We are building this library */
#      define TTKUSERINTERFACEBASE_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKUSERINTERFACEBASE_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKUSERINTERFACEBASE_NO_EXPORT
#    define TTKUSERINTERFACEBASE_NO_EXPORT 
#  endif
#endif

#ifndef TTKUSERINTERFACEBASE_DEPRECATED
#  define TTKUSERINTERFACEBASE_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKUSERINTERFACEBASE_DEPRECATED_EXPORT
#  define TTKUSERINTERFACEBASE_DEPRECATED_EXPORT TTKUSERINTERFACEBASE_EXPORT TTKUSERINTERFACEBASE_DEPRECATED
#endif

#ifndef TTKUSERINTERFACEBASE_DEPRECATED_NO_EXPORT
#  define TTKUSERINTERFACEBASE_DEPRECATED_NO_EXPORT TTKUSERINTERFACEBASE_NO_EXPORT TTKUSERINTERFACEBASE_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKUSERINTERFACEBASE_NO_DEPRECATED
#    define TTKUSERINTERFACEBASE_NO_DEPRECATED
#  endif
#endif

#endif /* TTKUSERINTERFACEBASE_EXPORT_H */
