
#ifndef TTKOBJWRITER_EXPORT_H
#define TTKOBJWRITER_EXPORT_H

#ifdef TTKOBJWRITER_STATIC_DEFINE
#  define TTKOBJWRITER_EXPORT
#  define TTKOBJWRITER_NO_EXPORT
#else
#  ifndef TTKOBJWRITER_EXPORT
#    ifdef ttkOBJWriter_EXPORTS
        /* We are building this library */
#      define TTKOBJWRITER_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKOBJWRITER_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKOBJWRITER_NO_EXPORT
#    define TTKOBJWRITER_NO_EXPORT 
#  endif
#endif

#ifndef TTKOBJWRITER_DEPRECATED
#  define TTKOBJWRITER_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKOBJWRITER_DEPRECATED_EXPORT
#  define TTKOBJWRITER_DEPRECATED_EXPORT TTKOBJWRITER_EXPORT TTKOBJWRITER_DEPRECATED
#endif

#ifndef TTKOBJWRITER_DEPRECATED_NO_EXPORT
#  define TTKOBJWRITER_DEPRECATED_NO_EXPORT TTKOBJWRITER_NO_EXPORT TTKOBJWRITER_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKOBJWRITER_NO_DEPRECATED
#    define TTKOBJWRITER_NO_DEPRECATED
#  endif
#endif

#endif /* TTKOBJWRITER_EXPORT_H */
