
#ifndef TTKTEXTUREMAPFROMFIELD_EXPORT_H
#define TTKTEXTUREMAPFROMFIELD_EXPORT_H

#ifdef TTKTEXTUREMAPFROMFIELD_STATIC_DEFINE
#  define TTKTEXTUREMAPFROMFIELD_EXPORT
#  define TTKTEXTUREMAPFROMFIELD_NO_EXPORT
#else
#  ifndef TTKTEXTUREMAPFROMFIELD_EXPORT
#    ifdef ttkTextureMapFromField_EXPORTS
        /* We are building this library */
#      define TTKTEXTUREMAPFROMFIELD_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKTEXTUREMAPFROMFIELD_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKTEXTUREMAPFROMFIELD_NO_EXPORT
#    define TTKTEXTUREMAPFROMFIELD_NO_EXPORT 
#  endif
#endif

#ifndef TTKTEXTUREMAPFROMFIELD_DEPRECATED
#  define TTKTEXTUREMAPFROMFIELD_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKTEXTUREMAPFROMFIELD_DEPRECATED_EXPORT
#  define TTKTEXTUREMAPFROMFIELD_DEPRECATED_EXPORT TTKTEXTUREMAPFROMFIELD_EXPORT TTKTEXTUREMAPFROMFIELD_DEPRECATED
#endif

#ifndef TTKTEXTUREMAPFROMFIELD_DEPRECATED_NO_EXPORT
#  define TTKTEXTUREMAPFROMFIELD_DEPRECATED_NO_EXPORT TTKTEXTUREMAPFROMFIELD_NO_EXPORT TTKTEXTUREMAPFROMFIELD_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKTEXTUREMAPFROMFIELD_NO_DEPRECATED
#    define TTKTEXTUREMAPFROMFIELD_NO_DEPRECATED
#  endif
#endif

#endif /* TTKTEXTUREMAPFROMFIELD_EXPORT_H */
