
#ifndef TTKTRIANGULATIONALGORITHM_EXPORT_H
#define TTKTRIANGULATIONALGORITHM_EXPORT_H

#ifdef TTKTRIANGULATIONALGORITHM_STATIC_DEFINE
#  define TTKTRIANGULATIONALGORITHM_EXPORT
#  define TTKTRIANGULATIONALGORITHM_NO_EXPORT
#else
#  ifndef TTKTRIANGULATIONALGORITHM_EXPORT
#    ifdef ttkTriangulationAlgorithm_EXPORTS
        /* We are building this library */
#      define TTKTRIANGULATIONALGORITHM_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKTRIANGULATIONALGORITHM_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKTRIANGULATIONALGORITHM_NO_EXPORT
#    define TTKTRIANGULATIONALGORITHM_NO_EXPORT 
#  endif
#endif

#ifndef TTKTRIANGULATIONALGORITHM_DEPRECATED
#  define TTKTRIANGULATIONALGORITHM_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKTRIANGULATIONALGORITHM_DEPRECATED_EXPORT
#  define TTKTRIANGULATIONALGORITHM_DEPRECATED_EXPORT TTKTRIANGULATIONALGORITHM_EXPORT TTKTRIANGULATIONALGORITHM_DEPRECATED
#endif

#ifndef TTKTRIANGULATIONALGORITHM_DEPRECATED_NO_EXPORT
#  define TTKTRIANGULATIONALGORITHM_DEPRECATED_NO_EXPORT TTKTRIANGULATIONALGORITHM_NO_EXPORT TTKTRIANGULATIONALGORITHM_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKTRIANGULATIONALGORITHM_NO_DEPRECATED
#    define TTKTRIANGULATIONALGORITHM_NO_DEPRECATED
#  endif
#endif

#endif /* TTKTRIANGULATIONALGORITHM_EXPORT_H */
