
#ifndef TTKSCALARFIELDNORMALIZER_EXPORT_H
#define TTKSCALARFIELDNORMALIZER_EXPORT_H

#ifdef TTKSCALARFIELDNORMALIZER_STATIC_DEFINE
#  define TTKSCALARFIELDNORMALIZER_EXPORT
#  define TTKSCALARFIELDNORMALIZER_NO_EXPORT
#else
#  ifndef TTKSCALARFIELDNORMALIZER_EXPORT
#    ifdef ttkScalarFieldNormalizer_EXPORTS
        /* We are building this library */
#      define TTKSCALARFIELDNORMALIZER_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKSCALARFIELDNORMALIZER_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKSCALARFIELDNORMALIZER_NO_EXPORT
#    define TTKSCALARFIELDNORMALIZER_NO_EXPORT 
#  endif
#endif

#ifndef TTKSCALARFIELDNORMALIZER_DEPRECATED
#  define TTKSCALARFIELDNORMALIZER_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKSCALARFIELDNORMALIZER_DEPRECATED_EXPORT
#  define TTKSCALARFIELDNORMALIZER_DEPRECATED_EXPORT TTKSCALARFIELDNORMALIZER_EXPORT TTKSCALARFIELDNORMALIZER_DEPRECATED
#endif

#ifndef TTKSCALARFIELDNORMALIZER_DEPRECATED_NO_EXPORT
#  define TTKSCALARFIELDNORMALIZER_DEPRECATED_NO_EXPORT TTKSCALARFIELDNORMALIZER_NO_EXPORT TTKSCALARFIELDNORMALIZER_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKSCALARFIELDNORMALIZER_NO_DEPRECATED
#    define TTKSCALARFIELDNORMALIZER_NO_DEPRECATED
#  endif
#endif

#endif /* TTKSCALARFIELDNORMALIZER_EXPORT_H */
