
#ifndef TTKEIGENFIELD_EXPORT_H
#define TTKEIGENFIELD_EXPORT_H

#ifdef TTKEIGENFIELD_STATIC_DEFINE
#  define TTKEIGENFIELD_EXPORT
#  define TTKEIGENFIELD_NO_EXPORT
#else
#  ifndef TTKEIGENFIELD_EXPORT
#    ifdef ttkEigenField_EXPORTS
        /* We are building this library */
#      define TTKEIGENFIELD_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKEIGENFIELD_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKEIGENFIELD_NO_EXPORT
#    define TTKEIGENFIELD_NO_EXPORT 
#  endif
#endif

#ifndef TTKEIGENFIELD_DEPRECATED
#  define TTKEIGENFIELD_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKEIGENFIELD_DEPRECATED_EXPORT
#  define TTKEIGENFIELD_DEPRECATED_EXPORT TTKEIGENFIELD_EXPORT TTKEIGENFIELD_DEPRECATED
#endif

#ifndef TTKEIGENFIELD_DEPRECATED_NO_EXPORT
#  define TTKEIGENFIELD_DEPRECATED_NO_EXPORT TTKEIGENFIELD_NO_EXPORT TTKEIGENFIELD_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKEIGENFIELD_NO_DEPRECATED
#    define TTKEIGENFIELD_NO_DEPRECATED
#  endif
#endif

#endif /* TTKEIGENFIELD_EXPORT_H */
