
#ifndef TTKIDENTIFYBYSCALARFIELD_EXPORT_H
#define TTKIDENTIFYBYSCALARFIELD_EXPORT_H

#ifdef TTKIDENTIFYBYSCALARFIELD_STATIC_DEFINE
#  define TTKIDENTIFYBYSCALARFIELD_EXPORT
#  define TTKIDENTIFYBYSCALARFIELD_NO_EXPORT
#else
#  ifndef TTKIDENTIFYBYSCALARFIELD_EXPORT
#    ifdef ttkIdentifyByScalarField_EXPORTS
        /* We are building this library */
#      define TTKIDENTIFYBYSCALARFIELD_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKIDENTIFYBYSCALARFIELD_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKIDENTIFYBYSCALARFIELD_NO_EXPORT
#    define TTKIDENTIFYBYSCALARFIELD_NO_EXPORT 
#  endif
#endif

#ifndef TTKIDENTIFYBYSCALARFIELD_DEPRECATED
#  define TTKIDENTIFYBYSCALARFIELD_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKIDENTIFYBYSCALARFIELD_DEPRECATED_EXPORT
#  define TTKIDENTIFYBYSCALARFIELD_DEPRECATED_EXPORT TTKIDENTIFYBYSCALARFIELD_EXPORT TTKIDENTIFYBYSCALARFIELD_DEPRECATED
#endif

#ifndef TTKIDENTIFYBYSCALARFIELD_DEPRECATED_NO_EXPORT
#  define TTKIDENTIFYBYSCALARFIELD_DEPRECATED_NO_EXPORT TTKIDENTIFYBYSCALARFIELD_NO_EXPORT TTKIDENTIFYBYSCALARFIELD_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKIDENTIFYBYSCALARFIELD_NO_DEPRECATED
#    define TTKIDENTIFYBYSCALARFIELD_NO_DEPRECATED
#  endif
#endif

#endif /* TTKIDENTIFYBYSCALARFIELD_EXPORT_H */
