
#ifndef TTKDIMENSIONREDUCTION_EXPORT_H
#define TTKDIMENSIONREDUCTION_EXPORT_H

#ifdef TTKDIMENSIONREDUCTION_STATIC_DEFINE
#  define TTKDIMENSIONREDUCTION_EXPORT
#  define TTKDIMENSIONREDUCTION_NO_EXPORT
#else
#  ifndef TTKDIMENSIONREDUCTION_EXPORT
#    ifdef ttkDimensionReduction_EXPORTS
        /* We are building this library */
#      define TTKDIMENSIONREDUCTION_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKDIMENSIONREDUCTION_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKDIMENSIONREDUCTION_NO_EXPORT
#    define TTKDIMENSIONREDUCTION_NO_EXPORT 
#  endif
#endif

#ifndef TTKDIMENSIONREDUCTION_DEPRECATED
#  define TTKDIMENSIONREDUCTION_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKDIMENSIONREDUCTION_DEPRECATED_EXPORT
#  define TTKDIMENSIONREDUCTION_DEPRECATED_EXPORT TTKDIMENSIONREDUCTION_EXPORT TTKDIMENSIONREDUCTION_DEPRECATED
#endif

#ifndef TTKDIMENSIONREDUCTION_DEPRECATED_NO_EXPORT
#  define TTKDIMENSIONREDUCTION_DEPRECATED_NO_EXPORT TTKDIMENSIONREDUCTION_NO_EXPORT TTKDIMENSIONREDUCTION_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKDIMENSIONREDUCTION_NO_DEPRECATED
#    define TTKDIMENSIONREDUCTION_NO_DEPRECATED
#  endif
#endif

#endif /* TTKDIMENSIONREDUCTION_EXPORT_H */
