
#ifndef TTKCONTOURAROUNDPOINT_EXPORT_H
#define TTKCONTOURAROUNDPOINT_EXPORT_H

#ifdef TTKCONTOURAROUNDPOINT_STATIC_DEFINE
#  define TTKCONTOURAROUNDPOINT_EXPORT
#  define TTKCONTOURAROUNDPOINT_NO_EXPORT
#else
#  ifndef TTKCONTOURAROUNDPOINT_EXPORT
#    ifdef ttkContourAroundPoint_EXPORTS
        /* We are building this library */
#      define TTKCONTOURAROUNDPOINT_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKCONTOURAROUNDPOINT_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKCONTOURAROUNDPOINT_NO_EXPORT
#    define TTKCONTOURAROUNDPOINT_NO_EXPORT 
#  endif
#endif

#ifndef TTKCONTOURAROUNDPOINT_DEPRECATED
#  define TTKCONTOURAROUNDPOINT_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKCONTOURAROUNDPOINT_DEPRECATED_EXPORT
#  define TTKCONTOURAROUNDPOINT_DEPRECATED_EXPORT TTKCONTOURAROUNDPOINT_EXPORT TTKCONTOURAROUNDPOINT_DEPRECATED
#endif

#ifndef TTKCONTOURAROUNDPOINT_DEPRECATED_NO_EXPORT
#  define TTKCONTOURAROUNDPOINT_DEPRECATED_NO_EXPORT TTKCONTOURAROUNDPOINT_NO_EXPORT TTKCONTOURAROUNDPOINT_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKCONTOURAROUNDPOINT_NO_DEPRECATED
#    define TTKCONTOURAROUNDPOINT_NO_DEPRECATED
#  endif
#endif

#endif /* TTKCONTOURAROUNDPOINT_EXPORT_H */
