
#ifndef TTKICOSPHERESFROMPOINTS_EXPORT_H
#define TTKICOSPHERESFROMPOINTS_EXPORT_H

#ifdef TTKICOSPHERESFROMPOINTS_STATIC_DEFINE
#  define TTKICOSPHERESFROMPOINTS_EXPORT
#  define TTKICOSPHERESFROMPOINTS_NO_EXPORT
#else
#  ifndef TTKICOSPHERESFROMPOINTS_EXPORT
#    ifdef ttkIcospheresFromPoints_EXPORTS
        /* We are building this library */
#      define TTKICOSPHERESFROMPOINTS_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKICOSPHERESFROMPOINTS_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKICOSPHERESFROMPOINTS_NO_EXPORT
#    define TTKICOSPHERESFROMPOINTS_NO_EXPORT 
#  endif
#endif

#ifndef TTKICOSPHERESFROMPOINTS_DEPRECATED
#  define TTKICOSPHERESFROMPOINTS_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKICOSPHERESFROMPOINTS_DEPRECATED_EXPORT
#  define TTKICOSPHERESFROMPOINTS_DEPRECATED_EXPORT TTKICOSPHERESFROMPOINTS_EXPORT TTKICOSPHERESFROMPOINTS_DEPRECATED
#endif

#ifndef TTKICOSPHERESFROMPOINTS_DEPRECATED_NO_EXPORT
#  define TTKICOSPHERESFROMPOINTS_DEPRECATED_NO_EXPORT TTKICOSPHERESFROMPOINTS_NO_EXPORT TTKICOSPHERESFROMPOINTS_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKICOSPHERESFROMPOINTS_NO_DEPRECATED
#    define TTKICOSPHERESFROMPOINTS_NO_DEPRECATED
#  endif
#endif

#endif /* TTKICOSPHERESFROMPOINTS_EXPORT_H */
