
#ifndef TTKEXTRACT_EXPORT_H
#define TTKEXTRACT_EXPORT_H

#ifdef TTKEXTRACT_STATIC_DEFINE
#  define TTKEXTRACT_EXPORT
#  define TTKEXTRACT_NO_EXPORT
#else
#  ifndef TTKEXTRACT_EXPORT
#    ifdef ttkExtract_EXPORTS
        /* We are building this library */
#      define TTKEXTRACT_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKEXTRACT_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKEXTRACT_NO_EXPORT
#    define TTKEXTRACT_NO_EXPORT 
#  endif
#endif

#ifndef TTKEXTRACT_DEPRECATED
#  define TTKEXTRACT_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKEXTRACT_DEPRECATED_EXPORT
#  define TTKEXTRACT_DEPRECATED_EXPORT TTKEXTRACT_EXPORT TTKEXTRACT_DEPRECATED
#endif

#ifndef TTKEXTRACT_DEPRECATED_NO_EXPORT
#  define TTKEXTRACT_DEPRECATED_NO_EXPORT TTKEXTRACT_NO_EXPORT TTKEXTRACT_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKEXTRACT_NO_DEPRECATED
#    define TTKEXTRACT_NO_DEPRECATED
#  endif
#endif

#endif /* TTKEXTRACT_EXPORT_H */
