
#ifndef TTKTRACKINGFROMPERSISTENCEDIAGRAMS_EXPORT_H
#define TTKTRACKINGFROMPERSISTENCEDIAGRAMS_EXPORT_H

#ifdef TTKTRACKINGFROMPERSISTENCEDIAGRAMS_STATIC_DEFINE
#  define TTKTRACKINGFROMPERSISTENCEDIAGRAMS_EXPORT
#  define TTKTRACKINGFROMPERSISTENCEDIAGRAMS_NO_EXPORT
#else
#  ifndef TTKTRACKINGFROMPERSISTENCEDIAGRAMS_EXPORT
#    ifdef ttkTrackingFromPersistenceDiagrams_EXPORTS
        /* We are building this library */
#      define TTKTRACKINGFROMPERSISTENCEDIAGRAMS_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKTRACKINGFROMPERSISTENCEDIAGRAMS_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKTRACKINGFROMPERSISTENCEDIAGRAMS_NO_EXPORT
#    define TTKTRACKINGFROMPERSISTENCEDIAGRAMS_NO_EXPORT 
#  endif
#endif

#ifndef TTKTRACKINGFROMPERSISTENCEDIAGRAMS_DEPRECATED
#  define TTKTRACKINGFROMPERSISTENCEDIAGRAMS_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKTRACKINGFROMPERSISTENCEDIAGRAMS_DEPRECATED_EXPORT
#  define TTKTRACKINGFROMPERSISTENCEDIAGRAMS_DEPRECATED_EXPORT TTKTRACKINGFROMPERSISTENCEDIAGRAMS_EXPORT TTKTRACKINGFROMPERSISTENCEDIAGRAMS_DEPRECATED
#endif

#ifndef TTKTRACKINGFROMPERSISTENCEDIAGRAMS_DEPRECATED_NO_EXPORT
#  define TTKTRACKINGFROMPERSISTENCEDIAGRAMS_DEPRECATED_NO_EXPORT TTKTRACKINGFROMPERSISTENCEDIAGRAMS_NO_EXPORT TTKTRACKINGFROMPERSISTENCEDIAGRAMS_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKTRACKINGFROMPERSISTENCEDIAGRAMS_NO_DEPRECATED
#    define TTKTRACKINGFROMPERSISTENCEDIAGRAMS_NO_DEPRECATED
#  endif
#endif

#endif /* TTKTRACKINGFROMPERSISTENCEDIAGRAMS_EXPORT_H */
