
#ifndef TTKICOSPHERE_EXPORT_H
#define TTKICOSPHERE_EXPORT_H

#ifdef TTKICOSPHERE_STATIC_DEFINE
#  define TTKICOSPHERE_EXPORT
#  define TTKICOSPHERE_NO_EXPORT
#else
#  ifndef TTKICOSPHERE_EXPORT
#    ifdef ttkIcosphere_EXPORTS
        /* We are building this library */
#      define TTKICOSPHERE_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKICOSPHERE_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKICOSPHERE_NO_EXPORT
#    define TTKICOSPHERE_NO_EXPORT 
#  endif
#endif

#ifndef TTKICOSPHERE_DEPRECATED
#  define TTKICOSPHERE_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKICOSPHERE_DEPRECATED_EXPORT
#  define TTKICOSPHERE_DEPRECATED_EXPORT TTKICOSPHERE_EXPORT TTKICOSPHERE_DEPRECATED
#endif

#ifndef TTKICOSPHERE_DEPRECATED_NO_EXPORT
#  define TTKICOSPHERE_DEPRECATED_NO_EXPORT TTKICOSPHERE_NO_EXPORT TTKICOSPHERE_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKICOSPHERE_NO_DEPRECATED
#    define TTKICOSPHERE_NO_DEPRECATED
#  endif
#endif

#endif /* TTKICOSPHERE_EXPORT_H */
