
#ifndef TTKWRLEXPORTER_EXPORT_H
#define TTKWRLEXPORTER_EXPORT_H

#ifdef TTKWRLEXPORTER_STATIC_DEFINE
#  define TTKWRLEXPORTER_EXPORT
#  define TTKWRLEXPORTER_NO_EXPORT
#else
#  ifndef TTKWRLEXPORTER_EXPORT
#    ifdef ttkWRLExporter_EXPORTS
        /* We are building this library */
#      define TTKWRLEXPORTER_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKWRLEXPORTER_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKWRLEXPORTER_NO_EXPORT
#    define TTKWRLEXPORTER_NO_EXPORT 
#  endif
#endif

#ifndef TTKWRLEXPORTER_DEPRECATED
#  define TTKWRLEXPORTER_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKWRLEXPORTER_DEPRECATED_EXPORT
#  define TTKWRLEXPORTER_DEPRECATED_EXPORT TTKWRLEXPORTER_EXPORT TTKWRLEXPORTER_DEPRECATED
#endif

#ifndef TTKWRLEXPORTER_DEPRECATED_NO_EXPORT
#  define TTKWRLEXPORTER_DEPRECATED_NO_EXPORT TTKWRLEXPORTER_NO_EXPORT TTKWRLEXPORTER_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKWRLEXPORTER_NO_DEPRECATED
#    define TTKWRLEXPORTER_NO_DEPRECATED
#  endif
#endif

#endif /* TTKWRLEXPORTER_EXPORT_H */
