
#ifndef TTKGEOMETRYSMOOTHER_EXPORT_H
#define TTKGEOMETRYSMOOTHER_EXPORT_H

#ifdef TTKGEOMETRYSMOOTHER_STATIC_DEFINE
#  define TTKGEOMETRYSMOOTHER_EXPORT
#  define TTKGEOMETRYSMOOTHER_NO_EXPORT
#else
#  ifndef TTKGEOMETRYSMOOTHER_EXPORT
#    ifdef ttkGeometrySmoother_EXPORTS
        /* We are building this library */
#      define TTKGEOMETRYSMOOTHER_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKGEOMETRYSMOOTHER_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKGEOMETRYSMOOTHER_NO_EXPORT
#    define TTKGEOMETRYSMOOTHER_NO_EXPORT 
#  endif
#endif

#ifndef TTKGEOMETRYSMOOTHER_DEPRECATED
#  define TTKGEOMETRYSMOOTHER_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKGEOMETRYSMOOTHER_DEPRECATED_EXPORT
#  define TTKGEOMETRYSMOOTHER_DEPRECATED_EXPORT TTKGEOMETRYSMOOTHER_EXPORT TTKGEOMETRYSMOOTHER_DEPRECATED
#endif

#ifndef TTKGEOMETRYSMOOTHER_DEPRECATED_NO_EXPORT
#  define TTKGEOMETRYSMOOTHER_DEPRECATED_NO_EXPORT TTKGEOMETRYSMOOTHER_NO_EXPORT TTKGEOMETRYSMOOTHER_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKGEOMETRYSMOOTHER_NO_DEPRECATED
#    define TTKGEOMETRYSMOOTHER_NO_DEPRECATED
#  endif
#endif

#endif /* TTKGEOMETRYSMOOTHER_EXPORT_H */
