
#ifndef TTKMESHSUBDIVISION_EXPORT_H
#define TTKMESHSUBDIVISION_EXPORT_H

#ifdef TTKMESHSUBDIVISION_STATIC_DEFINE
#  define TTKMESHSUBDIVISION_EXPORT
#  define TTKMESHSUBDIVISION_NO_EXPORT
#else
#  ifndef TTKMESHSUBDIVISION_EXPORT
#    ifdef ttkMeshSubdivision_EXPORTS
        /* We are building this library */
#      define TTKMESHSUBDIVISION_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKMESHSUBDIVISION_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKMESHSUBDIVISION_NO_EXPORT
#    define TTKMESHSUBDIVISION_NO_EXPORT 
#  endif
#endif

#ifndef TTKMESHSUBDIVISION_DEPRECATED
#  define TTKMESHSUBDIVISION_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKMESHSUBDIVISION_DEPRECATED_EXPORT
#  define TTKMESHSUBDIVISION_DEPRECATED_EXPORT TTKMESHSUBDIVISION_EXPORT TTKMESHSUBDIVISION_DEPRECATED
#endif

#ifndef TTKMESHSUBDIVISION_DEPRECATED_NO_EXPORT
#  define TTKMESHSUBDIVISION_DEPRECATED_NO_EXPORT TTKMESHSUBDIVISION_NO_EXPORT TTKMESHSUBDIVISION_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKMESHSUBDIVISION_NO_DEPRECATED
#    define TTKMESHSUBDIVISION_NO_DEPRECATED
#  endif
#endif

#endif /* TTKMESHSUBDIVISION_EXPORT_H */
