
#ifndef TTKPERIODICGRID_EXPORT_H
#define TTKPERIODICGRID_EXPORT_H

#ifdef TTKPERIODICGRID_STATIC_DEFINE
#  define TTKPERIODICGRID_EXPORT
#  define TTKPERIODICGRID_NO_EXPORT
#else
#  ifndef TTKPERIODICGRID_EXPORT
#    ifdef ttkPeriodicGrid_EXPORTS
        /* We are building this library */
#      define TTKPERIODICGRID_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKPERIODICGRID_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKPERIODICGRID_NO_EXPORT
#    define TTKPERIODICGRID_NO_EXPORT 
#  endif
#endif

#ifndef TTKPERIODICGRID_DEPRECATED
#  define TTKPERIODICGRID_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKPERIODICGRID_DEPRECATED_EXPORT
#  define TTKPERIODICGRID_DEPRECATED_EXPORT TTKPERIODICGRID_EXPORT TTKPERIODICGRID_DEPRECATED
#endif

#ifndef TTKPERIODICGRID_DEPRECATED_NO_EXPORT
#  define TTKPERIODICGRID_DEPRECATED_NO_EXPORT TTKPERIODICGRID_NO_EXPORT TTKPERIODICGRID_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKPERIODICGRID_NO_DEPRECATED
#    define TTKPERIODICGRID_NO_DEPRECATED
#  endif
#endif

#endif /* TTKPERIODICGRID_EXPORT_H */
