
#ifndef TTKDATASETINTERPOLATOR_EXPORT_H
#define TTKDATASETINTERPOLATOR_EXPORT_H

#ifdef TTKDATASETINTERPOLATOR_STATIC_DEFINE
#  define TTKDATASETINTERPOLATOR_EXPORT
#  define TTKDATASETINTERPOLATOR_NO_EXPORT
#else
#  ifndef TTKDATASETINTERPOLATOR_EXPORT
#    ifdef ttkDataSetInterpolator_EXPORTS
        /* We are building this library */
#      define TTKDATASETINTERPOLATOR_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKDATASETINTERPOLATOR_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKDATASETINTERPOLATOR_NO_EXPORT
#    define TTKDATASETINTERPOLATOR_NO_EXPORT 
#  endif
#endif

#ifndef TTKDATASETINTERPOLATOR_DEPRECATED
#  define TTKDATASETINTERPOLATOR_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKDATASETINTERPOLATOR_DEPRECATED_EXPORT
#  define TTKDATASETINTERPOLATOR_DEPRECATED_EXPORT TTKDATASETINTERPOLATOR_EXPORT TTKDATASETINTERPOLATOR_DEPRECATED
#endif

#ifndef TTKDATASETINTERPOLATOR_DEPRECATED_NO_EXPORT
#  define TTKDATASETINTERPOLATOR_DEPRECATED_NO_EXPORT TTKDATASETINTERPOLATOR_NO_EXPORT TTKDATASETINTERPOLATOR_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKDATASETINTERPOLATOR_NO_DEPRECATED
#    define TTKDATASETINTERPOLATOR_NO_DEPRECATED
#  endif
#endif

#endif /* TTKDATASETINTERPOLATOR_EXPORT_H */
