
#ifndef TTKREEBSPACE_EXPORT_H
#define TTKREEBSPACE_EXPORT_H

#ifdef TTKREEBSPACE_STATIC_DEFINE
#  define TTKREEBSPACE_EXPORT
#  define TTKREEBSPACE_NO_EXPORT
#else
#  ifndef TTKREEBSPACE_EXPORT
#    ifdef ttkReebSpace_EXPORTS
        /* We are building this library */
#      define TTKREEBSPACE_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKREEBSPACE_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKREEBSPACE_NO_EXPORT
#    define TTKREEBSPACE_NO_EXPORT 
#  endif
#endif

#ifndef TTKREEBSPACE_DEPRECATED
#  define TTKREEBSPACE_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKREEBSPACE_DEPRECATED_EXPORT
#  define TTKREEBSPACE_DEPRECATED_EXPORT TTKREEBSPACE_EXPORT TTKREEBSPACE_DEPRECATED
#endif

#ifndef TTKREEBSPACE_DEPRECATED_NO_EXPORT
#  define TTKREEBSPACE_DEPRECATED_NO_EXPORT TTKREEBSPACE_NO_EXPORT TTKREEBSPACE_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKREEBSPACE_NO_DEPRECATED
#    define TTKREEBSPACE_NO_DEPRECATED
#  endif
#endif

#endif /* TTKREEBSPACE_EXPORT_H */
