
#ifndef TTKPOINTSETTOCURVE_EXPORT_H
#define TTKPOINTSETTOCURVE_EXPORT_H

#ifdef TTKPOINTSETTOCURVE_STATIC_DEFINE
#  define TTKPOINTSETTOCURVE_EXPORT
#  define TTKPOINTSETTOCURVE_NO_EXPORT
#else
#  ifndef TTKPOINTSETTOCURVE_EXPORT
#    ifdef ttkPointSetToCurve_EXPORTS
        /* We are building this library */
#      define TTKPOINTSETTOCURVE_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKPOINTSETTOCURVE_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKPOINTSETTOCURVE_NO_EXPORT
#    define TTKPOINTSETTOCURVE_NO_EXPORT 
#  endif
#endif

#ifndef TTKPOINTSETTOCURVE_DEPRECATED
#  define TTKPOINTSETTOCURVE_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKPOINTSETTOCURVE_DEPRECATED_EXPORT
#  define TTKPOINTSETTOCURVE_DEPRECATED_EXPORT TTKPOINTSETTOCURVE_EXPORT TTKPOINTSETTOCURVE_DEPRECATED
#endif

#ifndef TTKPOINTSETTOCURVE_DEPRECATED_NO_EXPORT
#  define TTKPOINTSETTOCURVE_DEPRECATED_NO_EXPORT TTKPOINTSETTOCURVE_NO_EXPORT TTKPOINTSETTOCURVE_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKPOINTSETTOCURVE_NO_DEPRECATED
#    define TTKPOINTSETTOCURVE_NO_DEPRECATED
#  endif
#endif

#endif /* TTKPOINTSETTOCURVE_EXPORT_H */
