
#ifndef TTKTRIANGULATIONREQUEST_EXPORT_H
#define TTKTRIANGULATIONREQUEST_EXPORT_H

#ifdef TTKTRIANGULATIONREQUEST_STATIC_DEFINE
#  define TTKTRIANGULATIONREQUEST_EXPORT
#  define TTKTRIANGULATIONREQUEST_NO_EXPORT
#else
#  ifndef TTKTRIANGULATIONREQUEST_EXPORT
#    ifdef ttkTriangulationRequest_EXPORTS
        /* We are building this library */
#      define TTKTRIANGULATIONREQUEST_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKTRIANGULATIONREQUEST_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKTRIANGULATIONREQUEST_NO_EXPORT
#    define TTKTRIANGULATIONREQUEST_NO_EXPORT 
#  endif
#endif

#ifndef TTKTRIANGULATIONREQUEST_DEPRECATED
#  define TTKTRIANGULATIONREQUEST_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKTRIANGULATIONREQUEST_DEPRECATED_EXPORT
#  define TTKTRIANGULATIONREQUEST_DEPRECATED_EXPORT TTKTRIANGULATIONREQUEST_EXPORT TTKTRIANGULATIONREQUEST_DEPRECATED
#endif

#ifndef TTKTRIANGULATIONREQUEST_DEPRECATED_NO_EXPORT
#  define TTKTRIANGULATIONREQUEST_DEPRECATED_NO_EXPORT TTKTRIANGULATIONREQUEST_NO_EXPORT TTKTRIANGULATIONREQUEST_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKTRIANGULATIONREQUEST_NO_DEPRECATED
#    define TTKTRIANGULATIONREQUEST_NO_DEPRECATED
#  endif
#endif

#endif /* TTKTRIANGULATIONREQUEST_EXPORT_H */
