
#ifndef TTKARRAYPRECONDITIONING_EXPORT_H
#define TTKARRAYPRECONDITIONING_EXPORT_H

#ifdef TTKARRAYPRECONDITIONING_STATIC_DEFINE
#  define TTKARRAYPRECONDITIONING_EXPORT
#  define TTKARRAYPRECONDITIONING_NO_EXPORT
#else
#  ifndef TTKARRAYPRECONDITIONING_EXPORT
#    ifdef ttkArrayPreconditioning_EXPORTS
        /* We are building this library */
#      define TTKARRAYPRECONDITIONING_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKARRAYPRECONDITIONING_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKARRAYPRECONDITIONING_NO_EXPORT
#    define TTKARRAYPRECONDITIONING_NO_EXPORT 
#  endif
#endif

#ifndef TTKARRAYPRECONDITIONING_DEPRECATED
#  define TTKARRAYPRECONDITIONING_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKARRAYPRECONDITIONING_DEPRECATED_EXPORT
#  define TTKARRAYPRECONDITIONING_DEPRECATED_EXPORT TTKARRAYPRECONDITIONING_EXPORT TTKARRAYPRECONDITIONING_DEPRECATED
#endif

#ifndef TTKARRAYPRECONDITIONING_DEPRECATED_NO_EXPORT
#  define TTKARRAYPRECONDITIONING_DEPRECATED_NO_EXPORT TTKARRAYPRECONDITIONING_NO_EXPORT TTKARRAYPRECONDITIONING_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKARRAYPRECONDITIONING_NO_DEPRECATED
#    define TTKARRAYPRECONDITIONING_NO_DEPRECATED
#  endif
#endif

#endif /* TTKARRAYPRECONDITIONING_EXPORT_H */
