
#ifndef TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_EXPORT_H
#define TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_EXPORT_H

#ifdef TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_STATIC_DEFINE
#  define TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_EXPORT
#  define TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_NO_EXPORT
#else
#  ifndef TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_EXPORT
#    ifdef ttkDepthImageBasedGeometryApproximation_EXPORTS
        /* We are building this library */
#      define TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_NO_EXPORT
#    define TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_NO_EXPORT 
#  endif
#endif

#ifndef TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_DEPRECATED
#  define TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_DEPRECATED_EXPORT
#  define TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_DEPRECATED_EXPORT TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_EXPORT TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_DEPRECATED
#endif

#ifndef TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_DEPRECATED_NO_EXPORT
#  define TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_DEPRECATED_NO_EXPORT TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_NO_EXPORT TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_NO_DEPRECATED
#    define TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_NO_DEPRECATED
#  endif
#endif

#endif /* TTKDEPTHIMAGEBASEDGEOMETRYAPPROXIMATION_EXPORT_H */
