
#ifndef TTKFTMTREE_EXPORT_H
#define TTKFTMTREE_EXPORT_H

#ifdef TTKFTMTREE_STATIC_DEFINE
#  define TTKFTMTREE_EXPORT
#  define TTKFTMTREE_NO_EXPORT
#else
#  ifndef TTKFTMTREE_EXPORT
#    ifdef ttkFTMTree_EXPORTS
        /* We are building this library */
#      define TTKFTMTREE_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKFTMTREE_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKFTMTREE_NO_EXPORT
#    define TTKFTMTREE_NO_EXPORT 
#  endif
#endif

#ifndef TTKFTMTREE_DEPRECATED
#  define TTKFTMTREE_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKFTMTREE_DEPRECATED_EXPORT
#  define TTKFTMTREE_DEPRECATED_EXPORT TTKFTMTREE_EXPORT TTKFTMTREE_DEPRECATED
#endif

#ifndef TTKFTMTREE_DEPRECATED_NO_EXPORT
#  define TTKFTMTREE_DEPRECATED_NO_EXPORT TTKFTMTREE_NO_EXPORT TTKFTMTREE_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKFTMTREE_NO_DEPRECATED
#    define TTKFTMTREE_NO_DEPRECATED
#  endif
#endif

#endif /* TTKFTMTREE_EXPORT_H */
