
#ifndef TTKHELLOWORLD_EXPORT_H
#define TTKHELLOWORLD_EXPORT_H

#ifdef TTKHELLOWORLD_STATIC_DEFINE
#  define TTKHELLOWORLD_EXPORT
#  define TTKHELLOWORLD_NO_EXPORT
#else
#  ifndef TTKHELLOWORLD_EXPORT
#    ifdef ttkHelloWorld_EXPORTS
        /* We are building this library */
#      define TTKHELLOWORLD_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKHELLOWORLD_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKHELLOWORLD_NO_EXPORT
#    define TTKHELLOWORLD_NO_EXPORT 
#  endif
#endif

#ifndef TTKHELLOWORLD_DEPRECATED
#  define TTKHELLOWORLD_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKHELLOWORLD_DEPRECATED_EXPORT
#  define TTKHELLOWORLD_DEPRECATED_EXPORT TTKHELLOWORLD_EXPORT TTKHELLOWORLD_DEPRECATED
#endif

#ifndef TTKHELLOWORLD_DEPRECATED_NO_EXPORT
#  define TTKHELLOWORLD_DEPRECATED_NO_EXPORT TTKHELLOWORLD_NO_EXPORT TTKHELLOWORLD_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKHELLOWORLD_NO_DEPRECATED
#    define TTKHELLOWORLD_NO_DEPRECATED
#  endif
#endif

#endif /* TTKHELLOWORLD_EXPORT_H */
