
#ifndef TTKPROGRAMBASE_EXPORT_H
#define TTKPROGRAMBASE_EXPORT_H

#ifdef TTKPROGRAMBASE_STATIC_DEFINE
#  define TTKPROGRAMBASE_EXPORT
#  define TTKPROGRAMBASE_NO_EXPORT
#else
#  ifndef TTKPROGRAMBASE_EXPORT
#    ifdef ttkProgramBase_EXPORTS
        /* We are building this library */
#      define TTKPROGRAMBASE_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKPROGRAMBASE_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKPROGRAMBASE_NO_EXPORT
#    define TTKPROGRAMBASE_NO_EXPORT 
#  endif
#endif

#ifndef TTKPROGRAMBASE_DEPRECATED
#  define TTKPROGRAMBASE_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKPROGRAMBASE_DEPRECATED_EXPORT
#  define TTKPROGRAMBASE_DEPRECATED_EXPORT TTKPROGRAMBASE_EXPORT TTKPROGRAMBASE_DEPRECATED
#endif

#ifndef TTKPROGRAMBASE_DEPRECATED_NO_EXPORT
#  define TTKPROGRAMBASE_DEPRECATED_NO_EXPORT TTKPROGRAMBASE_NO_EXPORT TTKPROGRAMBASE_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKPROGRAMBASE_NO_DEPRECATED
#    define TTKPROGRAMBASE_NO_DEPRECATED
#  endif
#endif

#endif /* TTKPROGRAMBASE_EXPORT_H */
