
#ifndef TTKSPHEREFROMPOINT_EXPORT_H
#define TTKSPHEREFROMPOINT_EXPORT_H

#ifdef TTKSPHEREFROMPOINT_STATIC_DEFINE
#  define TTKSPHEREFROMPOINT_EXPORT
#  define TTKSPHEREFROMPOINT_NO_EXPORT
#else
#  ifndef TTKSPHEREFROMPOINT_EXPORT
#    ifdef ttkSphereFromPoint_EXPORTS
        /* We are building this library */
#      define TTKSPHEREFROMPOINT_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKSPHEREFROMPOINT_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKSPHEREFROMPOINT_NO_EXPORT
#    define TTKSPHEREFROMPOINT_NO_EXPORT 
#  endif
#endif

#ifndef TTKSPHEREFROMPOINT_DEPRECATED
#  define TTKSPHEREFROMPOINT_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKSPHEREFROMPOINT_DEPRECATED_EXPORT
#  define TTKSPHEREFROMPOINT_DEPRECATED_EXPORT TTKSPHEREFROMPOINT_EXPORT TTKSPHEREFROMPOINT_DEPRECATED
#endif

#ifndef TTKSPHEREFROMPOINT_DEPRECATED_NO_EXPORT
#  define TTKSPHEREFROMPOINT_DEPRECATED_NO_EXPORT TTKSPHEREFROMPOINT_NO_EXPORT TTKSPHEREFROMPOINT_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKSPHEREFROMPOINT_NO_DEPRECATED
#    define TTKSPHEREFROMPOINT_NO_DEPRECATED
#  endif
#endif

#endif /* TTKSPHEREFROMPOINT_EXPORT_H */
