
#ifndef TTKTOPOLOGICALCOMPRESSION_EXPORT_H
#define TTKTOPOLOGICALCOMPRESSION_EXPORT_H

#ifdef TTKTOPOLOGICALCOMPRESSION_STATIC_DEFINE
#  define TTKTOPOLOGICALCOMPRESSION_EXPORT
#  define TTKTOPOLOGICALCOMPRESSION_NO_EXPORT
#else
#  ifndef TTKTOPOLOGICALCOMPRESSION_EXPORT
#    ifdef ttkTopologicalCompression_EXPORTS
        /* We are building this library */
#      define TTKTOPOLOGICALCOMPRESSION_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKTOPOLOGICALCOMPRESSION_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKTOPOLOGICALCOMPRESSION_NO_EXPORT
#    define TTKTOPOLOGICALCOMPRESSION_NO_EXPORT 
#  endif
#endif

#ifndef TTKTOPOLOGICALCOMPRESSION_DEPRECATED
#  define TTKTOPOLOGICALCOMPRESSION_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKTOPOLOGICALCOMPRESSION_DEPRECATED_EXPORT
#  define TTKTOPOLOGICALCOMPRESSION_DEPRECATED_EXPORT TTKTOPOLOGICALCOMPRESSION_EXPORT TTKTOPOLOGICALCOMPRESSION_DEPRECATED
#endif

#ifndef TTKTOPOLOGICALCOMPRESSION_DEPRECATED_NO_EXPORT
#  define TTKTOPOLOGICALCOMPRESSION_DEPRECATED_NO_EXPORT TTKTOPOLOGICALCOMPRESSION_NO_EXPORT TTKTOPOLOGICALCOMPRESSION_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKTOPOLOGICALCOMPRESSION_NO_DEPRECATED
#    define TTKTOPOLOGICALCOMPRESSION_NO_DEPRECATED
#  endif
#endif

#endif /* TTKTOPOLOGICALCOMPRESSION_EXPORT_H */
