
#ifndef TTKMESHGRAPH_EXPORT_H
#define TTKMESHGRAPH_EXPORT_H

#ifdef TTKMESHGRAPH_STATIC_DEFINE
#  define TTKMESHGRAPH_EXPORT
#  define TTKMESHGRAPH_NO_EXPORT
#else
#  ifndef TTKMESHGRAPH_EXPORT
#    ifdef ttkMeshGraph_EXPORTS
        /* We are building this library */
#      define TTKMESHGRAPH_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKMESHGRAPH_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKMESHGRAPH_NO_EXPORT
#    define TTKMESHGRAPH_NO_EXPORT 
#  endif
#endif

#ifndef TTKMESHGRAPH_DEPRECATED
#  define TTKMESHGRAPH_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKMESHGRAPH_DEPRECATED_EXPORT
#  define TTKMESHGRAPH_DEPRECATED_EXPORT TTKMESHGRAPH_EXPORT TTKMESHGRAPH_DEPRECATED
#endif

#ifndef TTKMESHGRAPH_DEPRECATED_NO_EXPORT
#  define TTKMESHGRAPH_DEPRECATED_NO_EXPORT TTKMESHGRAPH_NO_EXPORT TTKMESHGRAPH_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKMESHGRAPH_NO_DEPRECATED
#    define TTKMESHGRAPH_NO_DEPRECATED
#  endif
#endif

#endif /* TTKMESHGRAPH_EXPORT_H */
