
#ifndef TTKINTEGRALLINES_EXPORT_H
#define TTKINTEGRALLINES_EXPORT_H

#ifdef TTKINTEGRALLINES_STATIC_DEFINE
#  define TTKINTEGRALLINES_EXPORT
#  define TTKINTEGRALLINES_NO_EXPORT
#else
#  ifndef TTKINTEGRALLINES_EXPORT
#    ifdef ttkIntegralLines_EXPORTS
        /* We are building this library */
#      define TTKINTEGRALLINES_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKINTEGRALLINES_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKINTEGRALLINES_NO_EXPORT
#    define TTKINTEGRALLINES_NO_EXPORT 
#  endif
#endif

#ifndef TTKINTEGRALLINES_DEPRECATED
#  define TTKINTEGRALLINES_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKINTEGRALLINES_DEPRECATED_EXPORT
#  define TTKINTEGRALLINES_DEPRECATED_EXPORT TTKINTEGRALLINES_EXPORT TTKINTEGRALLINES_DEPRECATED
#endif

#ifndef TTKINTEGRALLINES_DEPRECATED_NO_EXPORT
#  define TTKINTEGRALLINES_DEPRECATED_NO_EXPORT TTKINTEGRALLINES_NO_EXPORT TTKINTEGRALLINES_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKINTEGRALLINES_NO_DEPRECATED
#    define TTKINTEGRALLINES_NO_DEPRECATED
#  endif
#endif

#endif /* TTKINTEGRALLINES_EXPORT_H */
