
#ifndef TTKQUADRANGULATIONSUBDIVISION_EXPORT_H
#define TTKQUADRANGULATIONSUBDIVISION_EXPORT_H

#ifdef TTKQUADRANGULATIONSUBDIVISION_STATIC_DEFINE
#  define TTKQUADRANGULATIONSUBDIVISION_EXPORT
#  define TTKQUADRANGULATIONSUBDIVISION_NO_EXPORT
#else
#  ifndef TTKQUADRANGULATIONSUBDIVISION_EXPORT
#    ifdef ttkQuadrangulationSubdivision_EXPORTS
        /* We are building this library */
#      define TTKQUADRANGULATIONSUBDIVISION_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKQUADRANGULATIONSUBDIVISION_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKQUADRANGULATIONSUBDIVISION_NO_EXPORT
#    define TTKQUADRANGULATIONSUBDIVISION_NO_EXPORT 
#  endif
#endif

#ifndef TTKQUADRANGULATIONSUBDIVISION_DEPRECATED
#  define TTKQUADRANGULATIONSUBDIVISION_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKQUADRANGULATIONSUBDIVISION_DEPRECATED_EXPORT
#  define TTKQUADRANGULATIONSUBDIVISION_DEPRECATED_EXPORT TTKQUADRANGULATIONSUBDIVISION_EXPORT TTKQUADRANGULATIONSUBDIVISION_DEPRECATED
#endif

#ifndef TTKQUADRANGULATIONSUBDIVISION_DEPRECATED_NO_EXPORT
#  define TTKQUADRANGULATIONSUBDIVISION_DEPRECATED_NO_EXPORT TTKQUADRANGULATIONSUBDIVISION_NO_EXPORT TTKQUADRANGULATIONSUBDIVISION_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKQUADRANGULATIONSUBDIVISION_NO_DEPRECATED
#    define TTKQUADRANGULATIONSUBDIVISION_NO_DEPRECATED
#  endif
#endif

#endif /* TTKQUADRANGULATIONSUBDIVISION_EXPORT_H */
