
#ifndef TTKIDENTIFIERRANDOMIZER_EXPORT_H
#define TTKIDENTIFIERRANDOMIZER_EXPORT_H

#ifdef TTKIDENTIFIERRANDOMIZER_STATIC_DEFINE
#  define TTKIDENTIFIERRANDOMIZER_EXPORT
#  define TTKIDENTIFIERRANDOMIZER_NO_EXPORT
#else
#  ifndef TTKIDENTIFIERRANDOMIZER_EXPORT
#    ifdef ttkIdentifierRandomizer_EXPORTS
        /* We are building this library */
#      define TTKIDENTIFIERRANDOMIZER_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKIDENTIFIERRANDOMIZER_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKIDENTIFIERRANDOMIZER_NO_EXPORT
#    define TTKIDENTIFIERRANDOMIZER_NO_EXPORT 
#  endif
#endif

#ifndef TTKIDENTIFIERRANDOMIZER_DEPRECATED
#  define TTKIDENTIFIERRANDOMIZER_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKIDENTIFIERRANDOMIZER_DEPRECATED_EXPORT
#  define TTKIDENTIFIERRANDOMIZER_DEPRECATED_EXPORT TTKIDENTIFIERRANDOMIZER_EXPORT TTKIDENTIFIERRANDOMIZER_DEPRECATED
#endif

#ifndef TTKIDENTIFIERRANDOMIZER_DEPRECATED_NO_EXPORT
#  define TTKIDENTIFIERRANDOMIZER_DEPRECATED_NO_EXPORT TTKIDENTIFIERRANDOMIZER_NO_EXPORT TTKIDENTIFIERRANDOMIZER_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKIDENTIFIERRANDOMIZER_NO_DEPRECATED
#    define TTKIDENTIFIERRANDOMIZER_NO_DEPRECATED
#  endif
#endif

#endif /* TTKIDENTIFIERRANDOMIZER_EXPORT_H */
