
#ifndef TTKFOREACH_EXPORT_H
#define TTKFOREACH_EXPORT_H

#ifdef TTKFOREACH_STATIC_DEFINE
#  define TTKFOREACH_EXPORT
#  define TTKFOREACH_NO_EXPORT
#else
#  ifndef TTKFOREACH_EXPORT
#    ifdef ttkForEach_EXPORTS
        /* We are building this library */
#      define TTKFOREACH_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKFOREACH_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKFOREACH_NO_EXPORT
#    define TTKFOREACH_NO_EXPORT 
#  endif
#endif

#ifndef TTKFOREACH_DEPRECATED
#  define TTKFOREACH_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKFOREACH_DEPRECATED_EXPORT
#  define TTKFOREACH_DEPRECATED_EXPORT TTKFOREACH_EXPORT TTKFOREACH_DEPRECATED
#endif

#ifndef TTKFOREACH_DEPRECATED_NO_EXPORT
#  define TTKFOREACH_DEPRECATED_NO_EXPORT TTKFOREACH_NO_EXPORT TTKFOREACH_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKFOREACH_NO_DEPRECATED
#    define TTKFOREACH_NO_DEPRECATED
#  endif
#endif

#endif /* TTKFOREACH_EXPORT_H */
