
#ifndef TTKCONTOURFORESTS_EXPORT_H
#define TTKCONTOURFORESTS_EXPORT_H

#ifdef TTKCONTOURFORESTS_STATIC_DEFINE
#  define TTKCONTOURFORESTS_EXPORT
#  define TTKCONTOURFORESTS_NO_EXPORT
#else
#  ifndef TTKCONTOURFORESTS_EXPORT
#    ifdef ttkContourForests_EXPORTS
        /* We are building this library */
#      define TTKCONTOURFORESTS_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKCONTOURFORESTS_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKCONTOURFORESTS_NO_EXPORT
#    define TTKCONTOURFORESTS_NO_EXPORT 
#  endif
#endif

#ifndef TTKCONTOURFORESTS_DEPRECATED
#  define TTKCONTOURFORESTS_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKCONTOURFORESTS_DEPRECATED_EXPORT
#  define TTKCONTOURFORESTS_DEPRECATED_EXPORT TTKCONTOURFORESTS_EXPORT TTKCONTOURFORESTS_DEPRECATED
#endif

#ifndef TTKCONTOURFORESTS_DEPRECATED_NO_EXPORT
#  define TTKCONTOURFORESTS_DEPRECATED_NO_EXPORT TTKCONTOURFORESTS_NO_EXPORT TTKCONTOURFORESTS_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKCONTOURFORESTS_NO_DEPRECATED
#    define TTKCONTOURFORESTS_NO_DEPRECATED
#  endif
#endif

#endif /* TTKCONTOURFORESTS_EXPORT_H */
