
#ifndef TTKMANIFOLDCHECK_EXPORT_H
#define TTKMANIFOLDCHECK_EXPORT_H

#ifdef TTKMANIFOLDCHECK_STATIC_DEFINE
#  define TTKMANIFOLDCHECK_EXPORT
#  define TTKMANIFOLDCHECK_NO_EXPORT
#else
#  ifndef TTKMANIFOLDCHECK_EXPORT
#    ifdef ttkManifoldCheck_EXPORTS
        /* We are building this library */
#      define TTKMANIFOLDCHECK_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKMANIFOLDCHECK_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKMANIFOLDCHECK_NO_EXPORT
#    define TTKMANIFOLDCHECK_NO_EXPORT 
#  endif
#endif

#ifndef TTKMANIFOLDCHECK_DEPRECATED
#  define TTKMANIFOLDCHECK_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKMANIFOLDCHECK_DEPRECATED_EXPORT
#  define TTKMANIFOLDCHECK_DEPRECATED_EXPORT TTKMANIFOLDCHECK_EXPORT TTKMANIFOLDCHECK_DEPRECATED
#endif

#ifndef TTKMANIFOLDCHECK_DEPRECATED_NO_EXPORT
#  define TTKMANIFOLDCHECK_DEPRECATED_NO_EXPORT TTKMANIFOLDCHECK_NO_EXPORT TTKMANIFOLDCHECK_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKMANIFOLDCHECK_NO_DEPRECATED
#    define TTKMANIFOLDCHECK_NO_DEPRECATED
#  endif
#endif

#endif /* TTKMANIFOLDCHECK_EXPORT_H */
