
#ifndef TTKOFFWRITER_EXPORT_H
#define TTKOFFWRITER_EXPORT_H

#ifdef TTKOFFWRITER_STATIC_DEFINE
#  define TTKOFFWRITER_EXPORT
#  define TTKOFFWRITER_NO_EXPORT
#else
#  ifndef TTKOFFWRITER_EXPORT
#    ifdef ttkOFFWriter_EXPORTS
        /* We are building this library */
#      define TTKOFFWRITER_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKOFFWRITER_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKOFFWRITER_NO_EXPORT
#    define TTKOFFWRITER_NO_EXPORT 
#  endif
#endif

#ifndef TTKOFFWRITER_DEPRECATED
#  define TTKOFFWRITER_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKOFFWRITER_DEPRECATED_EXPORT
#  define TTKOFFWRITER_DEPRECATED_EXPORT TTKOFFWRITER_EXPORT TTKOFFWRITER_DEPRECATED
#endif

#ifndef TTKOFFWRITER_DEPRECATED_NO_EXPORT
#  define TTKOFFWRITER_DEPRECATED_NO_EXPORT TTKOFFWRITER_NO_EXPORT TTKOFFWRITER_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKOFFWRITER_NO_DEPRECATED
#    define TTKOFFWRITER_NO_DEPRECATED
#  endif
#endif

#endif /* TTKOFFWRITER_EXPORT_H */
