
#ifndef TTKSCALARFIELDCRITICALPOINTS_EXPORT_H
#define TTKSCALARFIELDCRITICALPOINTS_EXPORT_H

#ifdef TTKSCALARFIELDCRITICALPOINTS_STATIC_DEFINE
#  define TTKSCALARFIELDCRITICALPOINTS_EXPORT
#  define TTKSCALARFIELDCRITICALPOINTS_NO_EXPORT
#else
#  ifndef TTKSCALARFIELDCRITICALPOINTS_EXPORT
#    ifdef ttkScalarFieldCriticalPoints_EXPORTS
        /* We are building this library */
#      define TTKSCALARFIELDCRITICALPOINTS_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKSCALARFIELDCRITICALPOINTS_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKSCALARFIELDCRITICALPOINTS_NO_EXPORT
#    define TTKSCALARFIELDCRITICALPOINTS_NO_EXPORT 
#  endif
#endif

#ifndef TTKSCALARFIELDCRITICALPOINTS_DEPRECATED
#  define TTKSCALARFIELDCRITICALPOINTS_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKSCALARFIELDCRITICALPOINTS_DEPRECATED_EXPORT
#  define TTKSCALARFIELDCRITICALPOINTS_DEPRECATED_EXPORT TTKSCALARFIELDCRITICALPOINTS_EXPORT TTKSCALARFIELDCRITICALPOINTS_DEPRECATED
#endif

#ifndef TTKSCALARFIELDCRITICALPOINTS_DEPRECATED_NO_EXPORT
#  define TTKSCALARFIELDCRITICALPOINTS_DEPRECATED_NO_EXPORT TTKSCALARFIELDCRITICALPOINTS_NO_EXPORT TTKSCALARFIELDCRITICALPOINTS_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKSCALARFIELDCRITICALPOINTS_NO_DEPRECATED
#    define TTKSCALARFIELDCRITICALPOINTS_NO_DEPRECATED
#  endif
#endif

#endif /* TTKSCALARFIELDCRITICALPOINTS_EXPORT_H */
