
#ifndef TTKALGORITHM_EXPORT_H
#define TTKALGORITHM_EXPORT_H

#ifdef TTKALGORITHM_STATIC_DEFINE
#  define TTKALGORITHM_EXPORT
#  define TTKALGORITHM_NO_EXPORT
#else
#  ifndef TTKALGORITHM_EXPORT
#    ifdef ttkAlgorithm_EXPORTS
        /* We are building this library */
#      define TTKALGORITHM_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKALGORITHM_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKALGORITHM_NO_EXPORT
#    define TTKALGORITHM_NO_EXPORT 
#  endif
#endif

#ifndef TTKALGORITHM_DEPRECATED
#  define TTKALGORITHM_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKALGORITHM_DEPRECATED_EXPORT
#  define TTKALGORITHM_DEPRECATED_EXPORT TTKALGORITHM_EXPORT TTKALGORITHM_DEPRECATED
#endif

#ifndef TTKALGORITHM_DEPRECATED_NO_EXPORT
#  define TTKALGORITHM_DEPRECATED_NO_EXPORT TTKALGORITHM_NO_EXPORT TTKALGORITHM_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKALGORITHM_NO_DEPRECATED
#    define TTKALGORITHM_NO_DEPRECATED
#  endif
#endif

#endif /* TTKALGORITHM_EXPORT_H */
