
#ifndef TTKOFFREADER_EXPORT_H
#define TTKOFFREADER_EXPORT_H

#ifdef TTKOFFREADER_STATIC_DEFINE
#  define TTKOFFREADER_EXPORT
#  define TTKOFFREADER_NO_EXPORT
#else
#  ifndef TTKOFFREADER_EXPORT
#    ifdef ttkOFFReader_EXPORTS
        /* We are building this library */
#      define TTKOFFREADER_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKOFFREADER_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKOFFREADER_NO_EXPORT
#    define TTKOFFREADER_NO_EXPORT 
#  endif
#endif

#ifndef TTKOFFREADER_DEPRECATED
#  define TTKOFFREADER_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKOFFREADER_DEPRECATED_EXPORT
#  define TTKOFFREADER_DEPRECATED_EXPORT TTKOFFREADER_EXPORT TTKOFFREADER_DEPRECATED
#endif

#ifndef TTKOFFREADER_DEPRECATED_NO_EXPORT
#  define TTKOFFREADER_DEPRECATED_NO_EXPORT TTKOFFREADER_NO_EXPORT TTKOFFREADER_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKOFFREADER_NO_DEPRECATED
#    define TTKOFFREADER_NO_DEPRECATED
#  endif
#endif

#endif /* TTKOFFREADER_EXPORT_H */
