
#ifndef TTKCINEMADARKROOM_EXPORT_H
#define TTKCINEMADARKROOM_EXPORT_H

#ifdef TTKCINEMADARKROOM_STATIC_DEFINE
#  define TTKCINEMADARKROOM_EXPORT
#  define TTKCINEMADARKROOM_NO_EXPORT
#else
#  ifndef TTKCINEMADARKROOM_EXPORT
#    ifdef ttkCinemaDarkroom_EXPORTS
        /* We are building this library */
#      define TTKCINEMADARKROOM_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKCINEMADARKROOM_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKCINEMADARKROOM_NO_EXPORT
#    define TTKCINEMADARKROOM_NO_EXPORT 
#  endif
#endif

#ifndef TTKCINEMADARKROOM_DEPRECATED
#  define TTKCINEMADARKROOM_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKCINEMADARKROOM_DEPRECATED_EXPORT
#  define TTKCINEMADARKROOM_DEPRECATED_EXPORT TTKCINEMADARKROOM_EXPORT TTKCINEMADARKROOM_DEPRECATED
#endif

#ifndef TTKCINEMADARKROOM_DEPRECATED_NO_EXPORT
#  define TTKCINEMADARKROOM_DEPRECATED_NO_EXPORT TTKCINEMADARKROOM_NO_EXPORT TTKCINEMADARKROOM_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKCINEMADARKROOM_NO_DEPRECATED
#    define TTKCINEMADARKROOM_NO_DEPRECATED
#  endif
#endif

#endif /* TTKCINEMADARKROOM_EXPORT_H */
