
#ifndef TTKCONTINUOUSSCATTERPLOT_EXPORT_H
#define TTKCONTINUOUSSCATTERPLOT_EXPORT_H

#ifdef TTKCONTINUOUSSCATTERPLOT_STATIC_DEFINE
#  define TTKCONTINUOUSSCATTERPLOT_EXPORT
#  define TTKCONTINUOUSSCATTERPLOT_NO_EXPORT
#else
#  ifndef TTKCONTINUOUSSCATTERPLOT_EXPORT
#    ifdef ttkContinuousScatterPlot_EXPORTS
        /* We are building this library */
#      define TTKCONTINUOUSSCATTERPLOT_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKCONTINUOUSSCATTERPLOT_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKCONTINUOUSSCATTERPLOT_NO_EXPORT
#    define TTKCONTINUOUSSCATTERPLOT_NO_EXPORT 
#  endif
#endif

#ifndef TTKCONTINUOUSSCATTERPLOT_DEPRECATED
#  define TTKCONTINUOUSSCATTERPLOT_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKCONTINUOUSSCATTERPLOT_DEPRECATED_EXPORT
#  define TTKCONTINUOUSSCATTERPLOT_DEPRECATED_EXPORT TTKCONTINUOUSSCATTERPLOT_EXPORT TTKCONTINUOUSSCATTERPLOT_DEPRECATED
#endif

#ifndef TTKCONTINUOUSSCATTERPLOT_DEPRECATED_NO_EXPORT
#  define TTKCONTINUOUSSCATTERPLOT_DEPRECATED_NO_EXPORT TTKCONTINUOUSSCATTERPLOT_NO_EXPORT TTKCONTINUOUSSCATTERPLOT_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKCONTINUOUSSCATTERPLOT_NO_DEPRECATED
#    define TTKCONTINUOUSSCATTERPLOT_NO_DEPRECATED
#  endif
#endif

#endif /* TTKCONTINUOUSSCATTERPLOT_EXPORT_H */
