
#ifndef TTKIDENTIFIERS_EXPORT_H
#define TTKIDENTIFIERS_EXPORT_H

#ifdef TTKIDENTIFIERS_STATIC_DEFINE
#  define TTKIDENTIFIERS_EXPORT
#  define TTKIDENTIFIERS_NO_EXPORT
#else
#  ifndef TTKIDENTIFIERS_EXPORT
#    ifdef ttkIdentifiers_EXPORTS
        /* We are building this library */
#      define TTKIDENTIFIERS_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKIDENTIFIERS_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKIDENTIFIERS_NO_EXPORT
#    define TTKIDENTIFIERS_NO_EXPORT 
#  endif
#endif

#ifndef TTKIDENTIFIERS_DEPRECATED
#  define TTKIDENTIFIERS_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKIDENTIFIERS_DEPRECATED_EXPORT
#  define TTKIDENTIFIERS_DEPRECATED_EXPORT TTKIDENTIFIERS_EXPORT TTKIDENTIFIERS_DEPRECATED
#endif

#ifndef TTKIDENTIFIERS_DEPRECATED_NO_EXPORT
#  define TTKIDENTIFIERS_DEPRECATED_NO_EXPORT TTKIDENTIFIERS_NO_EXPORT TTKIDENTIFIERS_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKIDENTIFIERS_NO_DEPRECATED
#    define TTKIDENTIFIERS_NO_DEPRECATED
#  endif
#endif

#endif /* TTKIDENTIFIERS_EXPORT_H */
