
#ifndef TTKPOINTDATASELECTOR_EXPORT_H
#define TTKPOINTDATASELECTOR_EXPORT_H

#ifdef TTKPOINTDATASELECTOR_STATIC_DEFINE
#  define TTKPOINTDATASELECTOR_EXPORT
#  define TTKPOINTDATASELECTOR_NO_EXPORT
#else
#  ifndef TTKPOINTDATASELECTOR_EXPORT
#    ifdef ttkPointDataSelector_EXPORTS
        /* We are building this library */
#      define TTKPOINTDATASELECTOR_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKPOINTDATASELECTOR_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKPOINTDATASELECTOR_NO_EXPORT
#    define TTKPOINTDATASELECTOR_NO_EXPORT 
#  endif
#endif

#ifndef TTKPOINTDATASELECTOR_DEPRECATED
#  define TTKPOINTDATASELECTOR_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKPOINTDATASELECTOR_DEPRECATED_EXPORT
#  define TTKPOINTDATASELECTOR_DEPRECATED_EXPORT TTKPOINTDATASELECTOR_EXPORT TTKPOINTDATASELECTOR_DEPRECATED
#endif

#ifndef TTKPOINTDATASELECTOR_DEPRECATED_NO_EXPORT
#  define TTKPOINTDATASELECTOR_DEPRECATED_NO_EXPORT TTKPOINTDATASELECTOR_NO_EXPORT TTKPOINTDATASELECTOR_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKPOINTDATASELECTOR_NO_DEPRECATED
#    define TTKPOINTDATASELECTOR_NO_DEPRECATED
#  endif
#endif

#endif /* TTKPOINTDATASELECTOR_EXPORT_H */
