
#ifndef TTKCOMPONENTSIZE_EXPORT_H
#define TTKCOMPONENTSIZE_EXPORT_H

#ifdef TTKCOMPONENTSIZE_STATIC_DEFINE
#  define TTKCOMPONENTSIZE_EXPORT
#  define TTKCOMPONENTSIZE_NO_EXPORT
#else
#  ifndef TTKCOMPONENTSIZE_EXPORT
#    ifdef ttkComponentSize_EXPORTS
        /* We are building this library */
#      define TTKCOMPONENTSIZE_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKCOMPONENTSIZE_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKCOMPONENTSIZE_NO_EXPORT
#    define TTKCOMPONENTSIZE_NO_EXPORT 
#  endif
#endif

#ifndef TTKCOMPONENTSIZE_DEPRECATED
#  define TTKCOMPONENTSIZE_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKCOMPONENTSIZE_DEPRECATED_EXPORT
#  define TTKCOMPONENTSIZE_DEPRECATED_EXPORT TTKCOMPONENTSIZE_EXPORT TTKCOMPONENTSIZE_DEPRECATED
#endif

#ifndef TTKCOMPONENTSIZE_DEPRECATED_NO_EXPORT
#  define TTKCOMPONENTSIZE_DEPRECATED_NO_EXPORT TTKCOMPONENTSIZE_NO_EXPORT TTKCOMPONENTSIZE_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKCOMPONENTSIZE_NO_DEPRECATED
#    define TTKCOMPONENTSIZE_NO_DEPRECATED
#  endif
#endif

#endif /* TTKCOMPONENTSIZE_EXPORT_H */
