
#ifndef TTKENDFOR_EXPORT_H
#define TTKENDFOR_EXPORT_H

#ifdef TTKENDFOR_STATIC_DEFINE
#  define TTKENDFOR_EXPORT
#  define TTKENDFOR_NO_EXPORT
#else
#  ifndef TTKENDFOR_EXPORT
#    ifdef ttkEndFor_EXPORTS
        /* We are building this library */
#      define TTKENDFOR_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKENDFOR_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKENDFOR_NO_EXPORT
#    define TTKENDFOR_NO_EXPORT 
#  endif
#endif

#ifndef TTKENDFOR_DEPRECATED
#  define TTKENDFOR_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKENDFOR_DEPRECATED_EXPORT
#  define TTKENDFOR_DEPRECATED_EXPORT TTKENDFOR_EXPORT TTKENDFOR_DEPRECATED
#endif

#ifndef TTKENDFOR_DEPRECATED_NO_EXPORT
#  define TTKENDFOR_DEPRECATED_NO_EXPORT TTKENDFOR_NO_EXPORT TTKENDFOR_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKENDFOR_NO_DEPRECATED
#    define TTKENDFOR_NO_DEPRECATED
#  endif
#endif

#endif /* TTKENDFOR_EXPORT_H */
