
#ifndef TTKRANGEPOLYGON_EXPORT_H
#define TTKRANGEPOLYGON_EXPORT_H

#ifdef TTKRANGEPOLYGON_STATIC_DEFINE
#  define TTKRANGEPOLYGON_EXPORT
#  define TTKRANGEPOLYGON_NO_EXPORT
#else
#  ifndef TTKRANGEPOLYGON_EXPORT
#    ifdef ttkRangePolygon_EXPORTS
        /* We are building this library */
#      define TTKRANGEPOLYGON_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKRANGEPOLYGON_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKRANGEPOLYGON_NO_EXPORT
#    define TTKRANGEPOLYGON_NO_EXPORT 
#  endif
#endif

#ifndef TTKRANGEPOLYGON_DEPRECATED
#  define TTKRANGEPOLYGON_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKRANGEPOLYGON_DEPRECATED_EXPORT
#  define TTKRANGEPOLYGON_DEPRECATED_EXPORT TTKRANGEPOLYGON_EXPORT TTKRANGEPOLYGON_DEPRECATED
#endif

#ifndef TTKRANGEPOLYGON_DEPRECATED_NO_EXPORT
#  define TTKRANGEPOLYGON_DEPRECATED_NO_EXPORT TTKRANGEPOLYGON_NO_EXPORT TTKRANGEPOLYGON_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKRANGEPOLYGON_NO_DEPRECATED
#    define TTKRANGEPOLYGON_NO_DEPRECATED
#  endif
#endif

#endif /* TTKRANGEPOLYGON_EXPORT_H */
