
#ifndef TTKDISTANCEFIELD_EXPORT_H
#define TTKDISTANCEFIELD_EXPORT_H

#ifdef TTKDISTANCEFIELD_STATIC_DEFINE
#  define TTKDISTANCEFIELD_EXPORT
#  define TTKDISTANCEFIELD_NO_EXPORT
#else
#  ifndef TTKDISTANCEFIELD_EXPORT
#    ifdef ttkDistanceField_EXPORTS
        /* We are building this library */
#      define TTKDISTANCEFIELD_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKDISTANCEFIELD_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKDISTANCEFIELD_NO_EXPORT
#    define TTKDISTANCEFIELD_NO_EXPORT 
#  endif
#endif

#ifndef TTKDISTANCEFIELD_DEPRECATED
#  define TTKDISTANCEFIELD_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKDISTANCEFIELD_DEPRECATED_EXPORT
#  define TTKDISTANCEFIELD_DEPRECATED_EXPORT TTKDISTANCEFIELD_EXPORT TTKDISTANCEFIELD_DEPRECATED
#endif

#ifndef TTKDISTANCEFIELD_DEPRECATED_NO_EXPORT
#  define TTKDISTANCEFIELD_DEPRECATED_NO_EXPORT TTKDISTANCEFIELD_NO_EXPORT TTKDISTANCEFIELD_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKDISTANCEFIELD_NO_DEPRECATED
#    define TTKDISTANCEFIELD_NO_DEPRECATED
#  endif
#endif

#endif /* TTKDISTANCEFIELD_EXPORT_H */
