
#ifndef TTKMORSESMALECOMPLEX_EXPORT_H
#define TTKMORSESMALECOMPLEX_EXPORT_H

#ifdef TTKMORSESMALECOMPLEX_STATIC_DEFINE
#  define TTKMORSESMALECOMPLEX_EXPORT
#  define TTKMORSESMALECOMPLEX_NO_EXPORT
#else
#  ifndef TTKMORSESMALECOMPLEX_EXPORT
#    ifdef ttkMorseSmaleComplex_EXPORTS
        /* We are building this library */
#      define TTKMORSESMALECOMPLEX_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKMORSESMALECOMPLEX_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKMORSESMALECOMPLEX_NO_EXPORT
#    define TTKMORSESMALECOMPLEX_NO_EXPORT 
#  endif
#endif

#ifndef TTKMORSESMALECOMPLEX_DEPRECATED
#  define TTKMORSESMALECOMPLEX_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKMORSESMALECOMPLEX_DEPRECATED_EXPORT
#  define TTKMORSESMALECOMPLEX_DEPRECATED_EXPORT TTKMORSESMALECOMPLEX_EXPORT TTKMORSESMALECOMPLEX_DEPRECATED
#endif

#ifndef TTKMORSESMALECOMPLEX_DEPRECATED_NO_EXPORT
#  define TTKMORSESMALECOMPLEX_DEPRECATED_NO_EXPORT TTKMORSESMALECOMPLEX_NO_EXPORT TTKMORSESMALECOMPLEX_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKMORSESMALECOMPLEX_NO_DEPRECATED
#    define TTKMORSESMALECOMPLEX_NO_DEPRECATED
#  endif
#endif

#endif /* TTKMORSESMALECOMPLEX_EXPORT_H */
