
#ifndef TTKICOSPHEREFROMOBJECT_EXPORT_H
#define TTKICOSPHEREFROMOBJECT_EXPORT_H

#ifdef TTKICOSPHEREFROMOBJECT_STATIC_DEFINE
#  define TTKICOSPHEREFROMOBJECT_EXPORT
#  define TTKICOSPHEREFROMOBJECT_NO_EXPORT
#else
#  ifndef TTKICOSPHEREFROMOBJECT_EXPORT
#    ifdef ttkIcosphereFromObject_EXPORTS
        /* We are building this library */
#      define TTKICOSPHEREFROMOBJECT_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKICOSPHEREFROMOBJECT_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKICOSPHEREFROMOBJECT_NO_EXPORT
#    define TTKICOSPHEREFROMOBJECT_NO_EXPORT 
#  endif
#endif

#ifndef TTKICOSPHEREFROMOBJECT_DEPRECATED
#  define TTKICOSPHEREFROMOBJECT_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKICOSPHEREFROMOBJECT_DEPRECATED_EXPORT
#  define TTKICOSPHEREFROMOBJECT_DEPRECATED_EXPORT TTKICOSPHEREFROMOBJECT_EXPORT TTKICOSPHEREFROMOBJECT_DEPRECATED
#endif

#ifndef TTKICOSPHEREFROMOBJECT_DEPRECATED_NO_EXPORT
#  define TTKICOSPHEREFROMOBJECT_DEPRECATED_NO_EXPORT TTKICOSPHEREFROMOBJECT_NO_EXPORT TTKICOSPHEREFROMOBJECT_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKICOSPHEREFROMOBJECT_NO_DEPRECATED
#    define TTKICOSPHEREFROMOBJECT_NO_DEPRECATED
#  endif
#endif

#endif /* TTKICOSPHEREFROMOBJECT_EXPORT_H */
