
#ifndef TTKLDISTANCEMATRIX_EXPORT_H
#define TTKLDISTANCEMATRIX_EXPORT_H

#ifdef TTKLDISTANCEMATRIX_STATIC_DEFINE
#  define TTKLDISTANCEMATRIX_EXPORT
#  define TTKLDISTANCEMATRIX_NO_EXPORT
#else
#  ifndef TTKLDISTANCEMATRIX_EXPORT
#    ifdef ttkLDistanceMatrix_EXPORTS
        /* We are building this library */
#      define TTKLDISTANCEMATRIX_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKLDISTANCEMATRIX_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKLDISTANCEMATRIX_NO_EXPORT
#    define TTKLDISTANCEMATRIX_NO_EXPORT 
#  endif
#endif

#ifndef TTKLDISTANCEMATRIX_DEPRECATED
#  define TTKLDISTANCEMATRIX_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKLDISTANCEMATRIX_DEPRECATED_EXPORT
#  define TTKLDISTANCEMATRIX_DEPRECATED_EXPORT TTKLDISTANCEMATRIX_EXPORT TTKLDISTANCEMATRIX_DEPRECATED
#endif

#ifndef TTKLDISTANCEMATRIX_DEPRECATED_NO_EXPORT
#  define TTKLDISTANCEMATRIX_DEPRECATED_NO_EXPORT TTKLDISTANCEMATRIX_NO_EXPORT TTKLDISTANCEMATRIX_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKLDISTANCEMATRIX_NO_DEPRECATED
#    define TTKLDISTANCEMATRIX_NO_DEPRECATED
#  endif
#endif

#endif /* TTKLDISTANCEMATRIX_EXPORT_H */
