
#ifndef TTKCINEMAWRITER_EXPORT_H
#define TTKCINEMAWRITER_EXPORT_H

#ifdef TTKCINEMAWRITER_STATIC_DEFINE
#  define TTKCINEMAWRITER_EXPORT
#  define TTKCINEMAWRITER_NO_EXPORT
#else
#  ifndef TTKCINEMAWRITER_EXPORT
#    ifdef ttkCinemaWriter_EXPORTS
        /* We are building this library */
#      define TTKCINEMAWRITER_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKCINEMAWRITER_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKCINEMAWRITER_NO_EXPORT
#    define TTKCINEMAWRITER_NO_EXPORT 
#  endif
#endif

#ifndef TTKCINEMAWRITER_DEPRECATED
#  define TTKCINEMAWRITER_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKCINEMAWRITER_DEPRECATED_EXPORT
#  define TTKCINEMAWRITER_DEPRECATED_EXPORT TTKCINEMAWRITER_EXPORT TTKCINEMAWRITER_DEPRECATED
#endif

#ifndef TTKCINEMAWRITER_DEPRECATED_NO_EXPORT
#  define TTKCINEMAWRITER_DEPRECATED_NO_EXPORT TTKCINEMAWRITER_NO_EXPORT TTKCINEMAWRITER_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKCINEMAWRITER_NO_DEPRECATED
#    define TTKCINEMAWRITER_NO_DEPRECATED
#  endif
#endif

#endif /* TTKCINEMAWRITER_EXPORT_H */
