
#ifndef TTKBLOCKAGGREGATOR_EXPORT_H
#define TTKBLOCKAGGREGATOR_EXPORT_H

#ifdef TTKBLOCKAGGREGATOR_STATIC_DEFINE
#  define TTKBLOCKAGGREGATOR_EXPORT
#  define TTKBLOCKAGGREGATOR_NO_EXPORT
#else
#  ifndef TTKBLOCKAGGREGATOR_EXPORT
#    ifdef ttkBlockAggregator_EXPORTS
        /* We are building this library */
#      define TTKBLOCKAGGREGATOR_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKBLOCKAGGREGATOR_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKBLOCKAGGREGATOR_NO_EXPORT
#    define TTKBLOCKAGGREGATOR_NO_EXPORT 
#  endif
#endif

#ifndef TTKBLOCKAGGREGATOR_DEPRECATED
#  define TTKBLOCKAGGREGATOR_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKBLOCKAGGREGATOR_DEPRECATED_EXPORT
#  define TTKBLOCKAGGREGATOR_DEPRECATED_EXPORT TTKBLOCKAGGREGATOR_EXPORT TTKBLOCKAGGREGATOR_DEPRECATED
#endif

#ifndef TTKBLOCKAGGREGATOR_DEPRECATED_NO_EXPORT
#  define TTKBLOCKAGGREGATOR_DEPRECATED_NO_EXPORT TTKBLOCKAGGREGATOR_NO_EXPORT TTKBLOCKAGGREGATOR_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKBLOCKAGGREGATOR_NO_DEPRECATED
#    define TTKBLOCKAGGREGATOR_NO_DEPRECATED
#  endif
#endif

#endif /* TTKBLOCKAGGREGATOR_EXPORT_H */
