
#ifndef TTKPERSISTENCEDIAGRAMCLUSTERING_EXPORT_H
#define TTKPERSISTENCEDIAGRAMCLUSTERING_EXPORT_H

#ifdef TTKPERSISTENCEDIAGRAMCLUSTERING_STATIC_DEFINE
#  define TTKPERSISTENCEDIAGRAMCLUSTERING_EXPORT
#  define TTKPERSISTENCEDIAGRAMCLUSTERING_NO_EXPORT
#else
#  ifndef TTKPERSISTENCEDIAGRAMCLUSTERING_EXPORT
#    ifdef ttkPersistenceDiagramClustering_EXPORTS
        /* We are building this library */
#      define TTKPERSISTENCEDIAGRAMCLUSTERING_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKPERSISTENCEDIAGRAMCLUSTERING_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKPERSISTENCEDIAGRAMCLUSTERING_NO_EXPORT
#    define TTKPERSISTENCEDIAGRAMCLUSTERING_NO_EXPORT 
#  endif
#endif

#ifndef TTKPERSISTENCEDIAGRAMCLUSTERING_DEPRECATED
#  define TTKPERSISTENCEDIAGRAMCLUSTERING_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKPERSISTENCEDIAGRAMCLUSTERING_DEPRECATED_EXPORT
#  define TTKPERSISTENCEDIAGRAMCLUSTERING_DEPRECATED_EXPORT TTKPERSISTENCEDIAGRAMCLUSTERING_EXPORT TTKPERSISTENCEDIAGRAMCLUSTERING_DEPRECATED
#endif

#ifndef TTKPERSISTENCEDIAGRAMCLUSTERING_DEPRECATED_NO_EXPORT
#  define TTKPERSISTENCEDIAGRAMCLUSTERING_DEPRECATED_NO_EXPORT TTKPERSISTENCEDIAGRAMCLUSTERING_NO_EXPORT TTKPERSISTENCEDIAGRAMCLUSTERING_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKPERSISTENCEDIAGRAMCLUSTERING_NO_DEPRECATED
#    define TTKPERSISTENCEDIAGRAMCLUSTERING_NO_DEPRECATED
#  endif
#endif

#endif /* TTKPERSISTENCEDIAGRAMCLUSTERING_EXPORT_H */
