
#ifndef TTKIMPORTEMBEDDINGFROMTABLE_EXPORT_H
#define TTKIMPORTEMBEDDINGFROMTABLE_EXPORT_H

#ifdef TTKIMPORTEMBEDDINGFROMTABLE_STATIC_DEFINE
#  define TTKIMPORTEMBEDDINGFROMTABLE_EXPORT
#  define TTKIMPORTEMBEDDINGFROMTABLE_NO_EXPORT
#else
#  ifndef TTKIMPORTEMBEDDINGFROMTABLE_EXPORT
#    ifdef ttkImportEmbeddingFromTable_EXPORTS
        /* We are building this library */
#      define TTKIMPORTEMBEDDINGFROMTABLE_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKIMPORTEMBEDDINGFROMTABLE_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKIMPORTEMBEDDINGFROMTABLE_NO_EXPORT
#    define TTKIMPORTEMBEDDINGFROMTABLE_NO_EXPORT 
#  endif
#endif

#ifndef TTKIMPORTEMBEDDINGFROMTABLE_DEPRECATED
#  define TTKIMPORTEMBEDDINGFROMTABLE_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKIMPORTEMBEDDINGFROMTABLE_DEPRECATED_EXPORT
#  define TTKIMPORTEMBEDDINGFROMTABLE_DEPRECATED_EXPORT TTKIMPORTEMBEDDINGFROMTABLE_EXPORT TTKIMPORTEMBEDDINGFROMTABLE_DEPRECATED
#endif

#ifndef TTKIMPORTEMBEDDINGFROMTABLE_DEPRECATED_NO_EXPORT
#  define TTKIMPORTEMBEDDINGFROMTABLE_DEPRECATED_NO_EXPORT TTKIMPORTEMBEDDINGFROMTABLE_NO_EXPORT TTKIMPORTEMBEDDINGFROMTABLE_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKIMPORTEMBEDDINGFROMTABLE_NO_DEPRECATED
#    define TTKIMPORTEMBEDDINGFROMTABLE_NO_DEPRECATED
#  endif
#endif

#endif /* TTKIMPORTEMBEDDINGFROMTABLE_EXPORT_H */
