
#ifndef TTKCINEMAIMAGING_EXPORT_H
#define TTKCINEMAIMAGING_EXPORT_H

#ifdef TTKCINEMAIMAGING_STATIC_DEFINE
#  define TTKCINEMAIMAGING_EXPORT
#  define TTKCINEMAIMAGING_NO_EXPORT
#else
#  ifndef TTKCINEMAIMAGING_EXPORT
#    ifdef ttkCinemaImaging_EXPORTS
        /* We are building this library */
#      define TTKCINEMAIMAGING_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKCINEMAIMAGING_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKCINEMAIMAGING_NO_EXPORT
#    define TTKCINEMAIMAGING_NO_EXPORT 
#  endif
#endif

#ifndef TTKCINEMAIMAGING_DEPRECATED
#  define TTKCINEMAIMAGING_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKCINEMAIMAGING_DEPRECATED_EXPORT
#  define TTKCINEMAIMAGING_DEPRECATED_EXPORT TTKCINEMAIMAGING_EXPORT TTKCINEMAIMAGING_DEPRECATED
#endif

#ifndef TTKCINEMAIMAGING_DEPRECATED_NO_EXPORT
#  define TTKCINEMAIMAGING_DEPRECATED_NO_EXPORT TTKCINEMAIMAGING_NO_EXPORT TTKCINEMAIMAGING_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKCINEMAIMAGING_NO_DEPRECATED
#    define TTKCINEMAIMAGING_NO_DEPRECATED
#  endif
#endif

#endif /* TTKCINEMAIMAGING_EXPORT_H */
