
#ifndef TTKGAUSSIANPOINTCLOUD_EXPORT_H
#define TTKGAUSSIANPOINTCLOUD_EXPORT_H

#ifdef TTKGAUSSIANPOINTCLOUD_STATIC_DEFINE
#  define TTKGAUSSIANPOINTCLOUD_EXPORT
#  define TTKGAUSSIANPOINTCLOUD_NO_EXPORT
#else
#  ifndef TTKGAUSSIANPOINTCLOUD_EXPORT
#    ifdef ttkGaussianPointCloud_EXPORTS
        /* We are building this library */
#      define TTKGAUSSIANPOINTCLOUD_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKGAUSSIANPOINTCLOUD_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKGAUSSIANPOINTCLOUD_NO_EXPORT
#    define TTKGAUSSIANPOINTCLOUD_NO_EXPORT 
#  endif
#endif

#ifndef TTKGAUSSIANPOINTCLOUD_DEPRECATED
#  define TTKGAUSSIANPOINTCLOUD_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKGAUSSIANPOINTCLOUD_DEPRECATED_EXPORT
#  define TTKGAUSSIANPOINTCLOUD_DEPRECATED_EXPORT TTKGAUSSIANPOINTCLOUD_EXPORT TTKGAUSSIANPOINTCLOUD_DEPRECATED
#endif

#ifndef TTKGAUSSIANPOINTCLOUD_DEPRECATED_NO_EXPORT
#  define TTKGAUSSIANPOINTCLOUD_DEPRECATED_NO_EXPORT TTKGAUSSIANPOINTCLOUD_NO_EXPORT TTKGAUSSIANPOINTCLOUD_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKGAUSSIANPOINTCLOUD_NO_DEPRECATED
#    define TTKGAUSSIANPOINTCLOUD_NO_DEPRECATED
#  endif
#endif

#endif /* TTKGAUSSIANPOINTCLOUD_EXPORT_H */
