
#ifndef TTKFTRGRAPH_EXPORT_H
#define TTKFTRGRAPH_EXPORT_H

#ifdef TTKFTRGRAPH_STATIC_DEFINE
#  define TTKFTRGRAPH_EXPORT
#  define TTKFTRGRAPH_NO_EXPORT
#else
#  ifndef TTKFTRGRAPH_EXPORT
#    ifdef ttkFTRGraph_EXPORTS
        /* We are building this library */
#      define TTKFTRGRAPH_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKFTRGRAPH_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKFTRGRAPH_NO_EXPORT
#    define TTKFTRGRAPH_NO_EXPORT 
#  endif
#endif

#ifndef TTKFTRGRAPH_DEPRECATED
#  define TTKFTRGRAPH_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKFTRGRAPH_DEPRECATED_EXPORT
#  define TTKFTRGRAPH_DEPRECATED_EXPORT TTKFTRGRAPH_EXPORT TTKFTRGRAPH_DEPRECATED
#endif

#ifndef TTKFTRGRAPH_DEPRECATED_NO_EXPORT
#  define TTKFTRGRAPH_DEPRECATED_NO_EXPORT TTKFTRGRAPH_NO_EXPORT TTKFTRGRAPH_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKFTRGRAPH_NO_DEPRECATED
#    define TTKFTRGRAPH_NO_DEPRECATED
#  endif
#endif

#endif /* TTKFTRGRAPH_EXPORT_H */
