
#ifndef TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_EXPORT_H
#define TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_EXPORT_H

#ifdef TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_STATIC_DEFINE
#  define TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_EXPORT
#  define TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_NO_EXPORT
#else
#  ifndef TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_EXPORT
#    ifdef ttkPersistenceDiagramDistanceMatrix_EXPORTS
        /* We are building this library */
#      define TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_NO_EXPORT
#    define TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_NO_EXPORT 
#  endif
#endif

#ifndef TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_DEPRECATED
#  define TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_DEPRECATED_EXPORT
#  define TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_DEPRECATED_EXPORT TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_EXPORT TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_DEPRECATED
#endif

#ifndef TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_DEPRECATED_NO_EXPORT
#  define TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_DEPRECATED_NO_EXPORT TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_NO_EXPORT TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_NO_DEPRECATED
#    define TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_NO_DEPRECATED
#  endif
#endif

#endif /* TTKPERSISTENCEDIAGRAMDISTANCEMATRIX_EXPORT_H */
