



python examples/python/example.py examples/data/inputData.vtu
IF %ERRORLEVEL% NEQ 0 exit /B 1
ttkMorseSmaleComplexCmd -i examples/data/inputData.vtu -o ""
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
