
#ifndef TTKPLANARGRAPHLAYOUT_EXPORT_H
#define TTKPLANARGRAPHLAYOUT_EXPORT_H

#ifdef TTKPLANARGRAPHLAYOUT_STATIC_DEFINE
#  define TTKPLANARGRAPHLAYOUT_EXPORT
#  define TTKPLANARGRAPHLAYOUT_NO_EXPORT
#else
#  ifndef TTKPLANARGRAPHLAYOUT_EXPORT
#    ifdef ttkPlanarGraphLayout_EXPORTS
        /* We are building this library */
#      define TTKPLANARGRAPHLAYOUT_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKPLANARGRAPHLAYOUT_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKPLANARGRAPHLAYOUT_NO_EXPORT
#    define TTKPLANARGRAPHLAYOUT_NO_EXPORT 
#  endif
#endif

#ifndef TTKPLANARGRAPHLAYOUT_DEPRECATED
#  define TTKPLANARGRAPHLAYOUT_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKPLANARGRAPHLAYOUT_DEPRECATED_EXPORT
#  define TTKPLANARGRAPHLAYOUT_DEPRECATED_EXPORT TTKPLANARGRAPHLAYOUT_EXPORT TTKPLANARGRAPHLAYOUT_DEPRECATED
#endif

#ifndef TTKPLANARGRAPHLAYOUT_DEPRECATED_NO_EXPORT
#  define TTKPLANARGRAPHLAYOUT_DEPRECATED_NO_EXPORT TTKPLANARGRAPHLAYOUT_NO_EXPORT TTKPLANARGRAPHLAYOUT_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKPLANARGRAPHLAYOUT_NO_DEPRECATED
#    define TTKPLANARGRAPHLAYOUT_NO_DEPRECATED
#  endif
#endif

#endif /* TTKPLANARGRAPHLAYOUT_EXPORT_H */
