
#ifndef TTKPROJECTIONFROMFIELD_EXPORT_H
#define TTKPROJECTIONFROMFIELD_EXPORT_H

#ifdef TTKPROJECTIONFROMFIELD_STATIC_DEFINE
#  define TTKPROJECTIONFROMFIELD_EXPORT
#  define TTKPROJECTIONFROMFIELD_NO_EXPORT
#else
#  ifndef TTKPROJECTIONFROMFIELD_EXPORT
#    ifdef ttkProjectionFromField_EXPORTS
        /* We are building this library */
#      define TTKPROJECTIONFROMFIELD_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKPROJECTIONFROMFIELD_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKPROJECTIONFROMFIELD_NO_EXPORT
#    define TTKPROJECTIONFROMFIELD_NO_EXPORT 
#  endif
#endif

#ifndef TTKPROJECTIONFROMFIELD_DEPRECATED
#  define TTKPROJECTIONFROMFIELD_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKPROJECTIONFROMFIELD_DEPRECATED_EXPORT
#  define TTKPROJECTIONFROMFIELD_DEPRECATED_EXPORT TTKPROJECTIONFROMFIELD_EXPORT TTKPROJECTIONFROMFIELD_DEPRECATED
#endif

#ifndef TTKPROJECTIONFROMFIELD_DEPRECATED_NO_EXPORT
#  define TTKPROJECTIONFROMFIELD_DEPRECATED_NO_EXPORT TTKPROJECTIONFROMFIELD_NO_EXPORT TTKPROJECTIONFROMFIELD_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKPROJECTIONFROMFIELD_NO_DEPRECATED
#    define TTKPROJECTIONFROMFIELD_NO_DEPRECATED
#  endif
#endif

#endif /* TTKPROJECTIONFROMFIELD_EXPORT_H */
