
#ifndef TTKLDISTANCE_EXPORT_H
#define TTKLDISTANCE_EXPORT_H

#ifdef TTKLDISTANCE_STATIC_DEFINE
#  define TTKLDISTANCE_EXPORT
#  define TTKLDISTANCE_NO_EXPORT
#else
#  ifndef TTKLDISTANCE_EXPORT
#    ifdef ttkLDistance_EXPORTS
        /* We are building this library */
#      define TTKLDISTANCE_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKLDISTANCE_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKLDISTANCE_NO_EXPORT
#    define TTKLDISTANCE_NO_EXPORT 
#  endif
#endif

#ifndef TTKLDISTANCE_DEPRECATED
#  define TTKLDISTANCE_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKLDISTANCE_DEPRECATED_EXPORT
#  define TTKLDISTANCE_DEPRECATED_EXPORT TTKLDISTANCE_EXPORT TTKLDISTANCE_DEPRECATED
#endif

#ifndef TTKLDISTANCE_DEPRECATED_NO_EXPORT
#  define TTKLDISTANCE_DEPRECATED_NO_EXPORT TTKLDISTANCE_NO_EXPORT TTKLDISTANCE_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKLDISTANCE_NO_DEPRECATED
#    define TTKLDISTANCE_NO_DEPRECATED
#  endif
#endif

#endif /* TTKLDISTANCE_EXPORT_H */
