
#ifndef TTKCONTOURTREEALIGNMENT_EXPORT_H
#define TTKCONTOURTREEALIGNMENT_EXPORT_H

#ifdef TTKCONTOURTREEALIGNMENT_STATIC_DEFINE
#  define TTKCONTOURTREEALIGNMENT_EXPORT
#  define TTKCONTOURTREEALIGNMENT_NO_EXPORT
#else
#  ifndef TTKCONTOURTREEALIGNMENT_EXPORT
#    ifdef ttkContourTreeAlignment_EXPORTS
        /* We are building this library */
#      define TTKCONTOURTREEALIGNMENT_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKCONTOURTREEALIGNMENT_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKCONTOURTREEALIGNMENT_NO_EXPORT
#    define TTKCONTOURTREEALIGNMENT_NO_EXPORT 
#  endif
#endif

#ifndef TTKCONTOURTREEALIGNMENT_DEPRECATED
#  define TTKCONTOURTREEALIGNMENT_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKCONTOURTREEALIGNMENT_DEPRECATED_EXPORT
#  define TTKCONTOURTREEALIGNMENT_DEPRECATED_EXPORT TTKCONTOURTREEALIGNMENT_EXPORT TTKCONTOURTREEALIGNMENT_DEPRECATED
#endif

#ifndef TTKCONTOURTREEALIGNMENT_DEPRECATED_NO_EXPORT
#  define TTKCONTOURTREEALIGNMENT_DEPRECATED_NO_EXPORT TTKCONTOURTREEALIGNMENT_NO_EXPORT TTKCONTOURTREEALIGNMENT_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKCONTOURTREEALIGNMENT_NO_DEPRECATED
#    define TTKCONTOURTREEALIGNMENT_NO_DEPRECATED
#  endif
#endif

#endif /* TTKCONTOURTREEALIGNMENT_EXPORT_H */
