
#ifndef TTKMANDATORYCRITICALPOINTS_EXPORT_H
#define TTKMANDATORYCRITICALPOINTS_EXPORT_H

#ifdef TTKMANDATORYCRITICALPOINTS_STATIC_DEFINE
#  define TTKMANDATORYCRITICALPOINTS_EXPORT
#  define TTKMANDATORYCRITICALPOINTS_NO_EXPORT
#else
#  ifndef TTKMANDATORYCRITICALPOINTS_EXPORT
#    ifdef ttkMandatoryCriticalPoints_EXPORTS
        /* We are building this library */
#      define TTKMANDATORYCRITICALPOINTS_EXPORT __declspec(dllexport)
#    else
        /* We are using this library */
#      define TTKMANDATORYCRITICALPOINTS_EXPORT __declspec(dllimport)
#    endif
#  endif

#  ifndef TTKMANDATORYCRITICALPOINTS_NO_EXPORT
#    define TTKMANDATORYCRITICALPOINTS_NO_EXPORT 
#  endif
#endif

#ifndef TTKMANDATORYCRITICALPOINTS_DEPRECATED
#  define TTKMANDATORYCRITICALPOINTS_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TTKMANDATORYCRITICALPOINTS_DEPRECATED_EXPORT
#  define TTKMANDATORYCRITICALPOINTS_DEPRECATED_EXPORT TTKMANDATORYCRITICALPOINTS_EXPORT TTKMANDATORYCRITICALPOINTS_DEPRECATED
#endif

#ifndef TTKMANDATORYCRITICALPOINTS_DEPRECATED_NO_EXPORT
#  define TTKMANDATORYCRITICALPOINTS_DEPRECATED_NO_EXPORT TTKMANDATORYCRITICALPOINTS_NO_EXPORT TTKMANDATORYCRITICALPOINTS_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TTKMANDATORYCRITICALPOINTS_NO_DEPRECATED
#    define TTKMANDATORYCRITICALPOINTS_NO_DEPRECATED
#  endif
#endif

#endif /* TTKMANDATORYCRITICALPOINTS_EXPORT_H */
